
//			ClearScreen.cpp


#ifndef CLEAR_SCREEN_VER
#define CLEAR_SCREEN_VER							1.0.2.0


#ifndef ENABLE_VIRTUAL_TERMINAL_PROCESSING
#define ENABLE_VIRTUAL_TERMINAL_PROCESSING			0x0004
#endif	//	ENABLE_VIRTUAL_TERMINAL_PROCESSING



int CLS_WriteConsoleA( LPCSTR Text )
{
	return WriteConsoleA( ConsoleHandle, Text, (DWORD) strlen( Text ), NULL, NULL );
}



BOOL CLS_WinTermCheck()
{
	wchar_t temp[SHORT_BUF_LEN]			= L"";


	GetEnvironmentVariable( L"WT_SESSION", temp, SHORT_BUF_LEN );

	if ( temp[0] )
		return TRUE;

	return FALSE;
}



int CountPromptNewLines()
{
	wchar_t temp[SHORT_BUF_LEN]			= L"";
	
	int k				= 0;


	GetEnvironmentVariable( L"Prompt", temp, SHORT_BUF_LEN );

	if ( temp[0] == '\0' )
		return 0;


	for ( wchar_t *ch = temp; *ch; ch++ ) {
		if ( *ch == '$' ) {
			ch++;

			if ( *ch == '_' )
				k++;
		}
	}

	return k;
}



int ClearScreen( KEYINFO24 *ki )
{
	CONSOLE_SCREEN_BUFFER_INFO BufferInfo;

	int ViewportHeight				= 0;
	int ViewportWidth				= 0;
	int NewTopRow					= 0;
	int ScrollLines					= 0;

	SMALL_RECT ViewportPosition		= { 0, 0, 0, 0 };
	COORD Origin					= { 0, 0 };
	CHAR_INFO ScrollFill			= { 0, 0 };

	DWORD OldConsoleMode			= 0;
	DWORD NewConsoleMode			= 0;


	ki->nKey			= 0;

	if ( NewKeyHook )
		ki->pszKey[0]	= '\0';


	memset( &BufferInfo, 0x00, sizeof( CONSOLE_SCREEN_BUFFER_INFO ) );

	if ( !GetConsoleScreenBufferInfo( ConsoleHandle, &BufferInfo ) )
		return 0;

	if ( ( BufferInfo.dwSize.X == 0 ) || ( BufferInfo.dwSize.Y == 0 ) )
		return 0;


	if ( CLS_WinTermCheck() ) {
		char temp[SHORT_BUF_LEN]			= "";

		ScrollLines = BufferInfo.dwCursorPosition.Y - CountPromptNewLines();

		if ( ScrollLines > 0 ) {
			GetConsoleMode( ConsoleHandle, &OldConsoleMode );
			NewConsoleMode = OldConsoleMode | ENABLE_VIRTUAL_TERMINAL_PROCESSING;

			if ( NewConsoleMode != OldConsoleMode )
				SetConsoleMode( ConsoleHandle, NewConsoleMode );

			CLS_WriteConsoleA( "\x1b[H" );

			sprintf_s( temp, SHORT_BUF_LEN, "\x1b[%dS", ScrollLines );
			CLS_WriteConsoleA( temp );

			ki->nHomeRow	-= ScrollLines;
			ki->nRow		-= ScrollLines;

			if ( NewConsoleMode != OldConsoleMode )
				SetConsoleMode( ConsoleHandle, OldConsoleMode );
		}

		return 0;
	}


	ViewportHeight	= BufferInfo.srWindow.Bottom - BufferInfo.srWindow.Top + 1;
	ViewportWidth	= BufferInfo.srWindow.Right - BufferInfo.srWindow.Left + 1;
	NewTopRow		= ki->nHomeRow - CountPromptNewLines();

	if ( (  NewTopRow + ViewportHeight ) <= BufferInfo.dwSize.Y ) {

		//		We have room at the bottom, so just move the viewport :

		ViewportPosition.Top	= NewTopRow;
		ViewportPosition.Bottom	= NewTopRow + ViewportHeight - 1;
		ViewportPosition.Left	= 0;
		ViewportPosition.Right	= ViewportWidth - 1;

		SetConsoleWindowInfo( ConsoleHandle, TRUE, &ViewportPosition );
	}
	else {

		//		We need to scroll the text in the buffer to make room :

		ScrollLines = ViewportHeight - ( BufferInfo.dwSize.Y - NewTopRow );

		ViewportPosition.Top			= ScrollLines;
		ViewportPosition.Bottom			= NewTopRow;
		ViewportPosition.Left			= 0;
		ViewportPosition.Right			= ViewportWidth - 1;

		Origin.X						= 0;
		Origin.Y						= 0;

		ScrollFill.Char.UnicodeChar		= 0x20;
		ScrollFill.Attributes			= BufferInfo.wAttributes;

		ScrollConsoleScreenBuffer( ConsoleHandle, &ViewportPosition, NULL, Origin, &ScrollFill );

		ki->nHomeRow	-= ScrollLines;
		ki->nRow		-= ScrollLines;
	}


	return 0;
}


#endif	//	CLEAR_SCREEN_VER


/*

	1.0.0.0		2022-05-03

	1.0.1.0		2022-06-29
				Adds a kludge to count the number of $_ tokens in %PROMPT, to
				better accomodate multiline prompts.

	1.0.2.0		2022-07-18
				Adds a kludge to sorta-kinda support Windows Terminal, which is less
				than perfectly compatible with the Windows console API.

*/
