
//		ConsoleMath.cpp	--		Automatically evaluate an expression beginning with a colon.


#ifndef CONSOLE_MATH_CPP
#define CONSOLE_MATH_CPP			1.0.1.0


const wchar_t CMath_VarName[]				= L"Ans";
const wchar_t CMath_HighlightVarName[]		= L"Highlight";
const wchar_t CMath_HighlightKeyName[]		= L"SOFTWARE\\JPPlugins";



int CMath_GetRegHighlightVal( HKEY RootKey, LPCTSTR KeyName, LPTSTR outString, DWORD buflen )
{
	if ( !RootKey || !KeyName || !outString || ( buflen < 128 ) )
		return -1;
	else if ( KeyName[0] == '\0' )
		return -1;


	HKEY KeyHandle			= NULL;
	DWORD ValSize			= 0;
	DWORD ValType			= 0;
	int rv					= 0;


	outString[0] = '\0';

	rv = RegOpenKeyEx( RootKey, KeyName, 0, KEY_READ, &KeyHandle );

	if ( rv == ERROR_SUCCESS ) {
		ValSize = buflen * sizeof( wchar_t );
		ValType = 0;

		rv = RegQueryValueEx( KeyHandle, CMath_HighlightVarName, NULL, &ValType, (BYTE*) outString, &ValSize );

		if ( rv == ERROR_SUCCESS ) {
			if ( ValType != REG_SZ )
				rv = ERROR_INVALID_DATA;
			else if ( outString[0] == '\0' )
				rv = ERROR_NO_DATA;
		}

		RegCloseKey( KeyHandle );
	}

	if ( rv )
		outString[0] = '\0';

	return rv;
}



BYTE CMath_PickHighlightColor( BYTE *Original )
{
	wchar_t temp[SHORT_BUF_LEN]			= L"";

	int fg = 0, bg = 0;
	unsigned int Normal = 0, Inverse = 0;
	int efg = -1, ebg = -1;

	const BYTE FirstColor[16]	= { 0x0f, 0x1f, 0x2f, 0x3f, 0x4f, 0x5f, 0x6f, 0x70, 0x8f, 0x9f, 0xa0, 0xb0, 0xcf, 0xd0, 0xe0, 0xf2 };
	const BYTE SecondColor[16]	= { 0x0e, 0x1b, 0x2b, 0x3e, 0x4b, 0x5b, 0x6b, 0x79, 0x8b, 0x9b, 0xa9, 0xb9, 0xc0, 0xde, 0xec, 0xf0 };


	GetAttribute( &Normal, &Inverse );
	fg = Normal & 0x0f;
	bg = ( Normal & 0xf0 ) >> 4;

	if ( Original )
		*Original = (BYTE) Normal & 0xff;

	if ( fg == bg )
		return 0x07;


	GetEnvironmentVariable( CMath_HighlightVarName, temp, SHORT_BUF_LEN );


	if ( temp[0] == '\0' ) {
		wchar_t KeyName[256] = L"";

		swprintf( KeyName, 256, L"%s\\%s", CMath_HighlightKeyName, PLUGIN_NAME );

		CMath_GetRegHighlightVal( HKEY_CURRENT_USER, KeyName, temp, SHORT_BUF_LEN );

		if ( temp[0] == '\0' )
			CMath_GetRegHighlightVal( HKEY_LOCAL_MACHINE, KeyName, temp, SHORT_BUF_LEN );

		if ( temp[0] == '\0' )
			CMath_GetRegHighlightVal( HKEY_CURRENT_USER, CMath_HighlightKeyName, temp, SHORT_BUF_LEN );

		if ( temp[0] == '\0' )
			CMath_GetRegHighlightVal( HKEY_LOCAL_MACHINE, CMath_HighlightKeyName, temp, SHORT_BUF_LEN );
	}


	if ( temp[0] ) {

		if ( !_wcsnicmp( temp, L"inv", 3 ) )
			ebg = ( Normal & 0x07 ), efg = bg;
		else if ( !_wcsnicmp( temp, L"0x", 2 ) ) {
			unsigned int t = 0;
			wchar_t *sep1 = NULL;

			_set_errno( 0 );
			t = wcstoul( temp + 2, &sep1, 16 );

			if ( !errno && ( *sep1 == '\0' ) )
				efg = t & 0x0f, ebg = ( t & 0xf0 ) >> 4;
		}
		else
			ParseColors( temp, &efg, &ebg );

		if ( ( ebg == 64 ) && ( efg >= 0 ) && ( efg <= 15 ) )		//  kludge:  values 1024 - 1039 treated as 0 - 15,
			ebg = -1;												//    but with default background color.

		if ( ( efg >= 0 ) && ( ebg >= 0 ) )
			if ( ( efg != ebg ) && ( ( efg != fg ) || ( ebg != bg ) ) )
				return ( ebg << 4 ) | efg;

		if ( ( efg >= 0 ) && ( ebg == -1 ) )
			if ( ( efg != fg ) && ( efg != bg ) )
				return ( bg << 4 ) | efg;

		return Normal;
	}

	if ( ( ( bg & 0x08 ) == 0 ) && ( ( fg & 0x08 ) == 0 ) && fg )
		return Normal | 0x08;

	if ( Normal == FirstColor[bg] )
		return SecondColor[bg];
	else
		return FirstColor[bg];
}



int DoConsoleMath( KEYINFO24 *ki )
{
	if ( !ki )
		return 0;

	if ( ki->pszLine[0] != ':' )
		return 0;

	if ( wcslen( ki->pszLine ) >= MAX_ARG_LEN )
		return 0;


	wchar_t Exp[MAX_ARG_LEN]				= L"";
	wchar_t Value[MAX_ARG_LEN]				= L"";

	BYTE OldColor		= 0;
	BYTE Highlight		= 0;
	int Row				= 0;
	int Column			= 0;
	int rv				= 0;


	wcscpy_s( Exp, MAX_ARG_LEN, ki->pszLine + 1 );
	TrimLine( Exp );

	if ( Exp[0] == '\0' )
		return 0;


	CrLf();

	wcscpy_s( Value, MAX_ARG_LEN, Exp );

	if ( wcschr( Value, '%' ) )
		if ( ExpandVariables( Value, 1 ) )
			goto Exit;

	if ( Evaluate( Value ) )
		goto Exit;


	Highlight = CMath_PickHighlightColor( &OldColor );
	
	Printf( L"%s = ", Exp );
	SetColors( Highlight );
	QPuts( Value );
	SetColors( OldColor );
	CrLf();
	SetEnvironmentVariable( CMath_VarName, Value );
	rv = 1;

Exit:

	//		Update the cursor position so the next prompt doesn't overwrite our output.
	//		Current versions of TCC don't need this, but TCC/LE does.

	GetAbsCursorPosition( &Row, &Column );
	ki->nHomeRow		= Row;
	ki->nHomeColumn		= 0;
	ki->nRow			= Row;
	ki->nColumn			= 0;
	return rv;
}


#endif	//	CONSOLE_MATH_CPP


/*

	1.0.0.0		2023-04-07

	1.0.1.0		2023-07-25
				Added the ability to read the Highlight variable from the registry.

*/
