
//			CopyLine.cpp


#ifndef COPYLINE_VERSION
#define COPYLINE_VERSION			1.0.1.0



BOOL CCL_AllBlanks( LPCTSTR String )
{
	if ( !String )
		return FALSE;


	for ( const wchar_t *ch = String; *ch; ch++ )
		if ( !iswspace( *ch ) )
			return FALSE;

	return TRUE;
}



int CopyCommandLine( KEYINFO24 *ki )
{
	if ( ki->pszLine == NULL )
		return 0;
	else if ( ki->pszLine[0] == '\0' )
		return 0;

	if ( CCL_AllBlanks( ki->pszLine ) )
		return 0;


	size_t len				= wcslen( ki->pszLine );
	HGLOBAL MemHandle		= NULL;
	wchar_t *Buffer			= NULL;
	int rv					= 1;


	MemHandle = GlobalAlloc( GHND, ( len + 1 ) * sizeof( wchar_t ) );

	if ( MemHandle ) {
		Buffer = (wchar_t*) GlobalLock( MemHandle );
		if ( Buffer ) {
			wcscpy_s( Buffer, len + 1, ki->pszLine );
			GlobalUnlock( MemHandle );
			rv = 0;
		}
	}

	if ( rv )
		goto Fail;


	rv = 2;

	if ( OpenClipboard( GetConsoleWindow() ) ) {
		if ( EmptyClipboard() )
			if ( SetClipboardData( CF_UNICODETEXT, MemHandle ) )
				rv = 0;

		CloseClipboard();
	}

	if ( rv )
		goto Fail;


	ki->nKey = 0;
	if ( NewKeyHook )
		ki->pszKey[0] = '\0';

	CopyChime();
	return 0;


Fail:

	if ( MemHandle )
		GlobalFree( MemHandle );

	ki->nKey = 0;
	if ( NewKeyHook )
		ki->pszKey[0] = '\0';

	ErrorBeep();
	return 0;
}



int CopyCommandWord( KEYINFO24 *ki, BOOL Cut )
{
	if ( ki->pszLine == NULL )
		return 0;
	else if ( ki->pszLine[0] == '\0' )
		return 0;

	if ( CCL_AllBlanks( ki->pszLine ) )
		return 0;


	wchar_t Word[MAX_ARG_LEN]		= L"";

	size_t WordStart		= 0;
	size_t WordLen			= 0;

	HGLOBAL MemHandle		= NULL;
	wchar_t *Buffer			= NULL;
	int rv					= 1;


	if ( !FindWordAtCursor( ki, &WordStart, &WordLen, Word, MAX_ARG_LEN ) )
		goto Fail;

	//	FmtMsgBox( 0, PLUGIN_NAME, L"Word = '%s'   WordLen = %I64u", Word, WordLen );


	MemHandle = GlobalAlloc( GHND, ( WordLen + 1 ) * sizeof( wchar_t ) );

	if ( MemHandle ) {
		Buffer = (wchar_t*) GlobalLock( MemHandle );
		if ( Buffer ) {
			wcscpy_s( Buffer, WordLen + 1, Word );
			GlobalUnlock( MemHandle );
			rv = 0;
		}
	}

	if ( rv )
		goto Fail;


	rv = 2;

	if ( OpenClipboard( GetConsoleWindow() ) ) {
		if ( EmptyClipboard() )
			if ( SetClipboardData( CF_UNICODETEXT, MemHandle ) )
				rv = 0;

		CloseClipboard();
	}

	if ( rv )
		goto Fail;


	if ( Cut ) {
		size_t len = WordLen;

		while ( iswspace( ki->pszLine[WordStart + len] ) )
			len++;

		CmdLineDelete( ki, WordStart, len );
	}


	ki->nKey = 0;
	if ( NewKeyHook )
		ki->pszKey[0] = '\0';

	if ( !Cut )
		CopyChime();

	return 0;


Fail:

	if ( MemHandle )
		GlobalFree( MemHandle );

	ki->nKey = 0;
	if ( NewKeyHook )
		ki->pszKey[0] = '\0';

	ErrorBeep();
	return 0;
}



#endif	//	COPYLINE_VERSION


/*

	1.0.0.0		2022-05-06

	1.0.1.0		2022-05-12
				Don't chime on a successful cut.  (The visible change to
				the command line should serve to indicate success.)

*/
