
//		EditKeys.cpp - Adds some new editing hotkeys to TCC.


#define UNICODE						1
#define _UNICODE					1

#define _CRT_RAND_S

#include <stdio.h>
#include <tchar.h>

#include <windows.h>

#define NEED_GDIPLUS

#ifdef NEED_GDIPLUS
#include <GdiPlus.h>
using namespace Gdiplus;
#endif	//	NEED_GDIPLUS

#include "PlugIn.h"
#include "TakeCmd.h"


#include "resource.h"


#define VER_MAJOR					0
#define VER_MINOR					99
#define VER_BUILD					12
#define VER_PATCH					2

#define PLUGIN_NAME					L"EditKeys"

#ifdef _WIN64
#define PLATFORM					L"x64"
#else
#define PLATFORM					L"x86"
#endif

#define MAX_ARG_LEN					8192
#define SHORT_BUF_LEN				1024
#define MAX_FILENAME_LEN			32768
#define MAX_CMD_LINE_LEN			65536
#define MAX_DOTS					256

#define ERRHANDLE					GetStdHandle( STD_ERROR_HANDLE )


#define MACRO_EXP( arg )			#arg
#define MACRO_STR( arg )			MACRO_EXP( arg )


#define MAX_KEYS					64
#define ERROR_BEEP_MAX				6
#define ERROR_BEEP_DEFAULT			2
#define COPY_BEEP_MAX				6
#define COPY_BEEP_DEFAULT			1

#define ACTION_SWAP_CHARS			1
#define ACTION_SWAP_ARGS			2
#define ACTION_INSERT_LAST_ARGS		3
#define ACTION_CLEAR_SCREEN			4
#define ACTION_MATH_WORD			5
#define ACTION_EXPAND_EXPR			6
#define ACTION_UNICODE_ENTRY		7
#define ACTION_COPY_LINE			8
#define ACTION_COPY_WORD			9
#define ACTION_CUT_WORD				10
#define ACTION_SWAP_CLIP_WORD		11
#define ACTION_CHANGE_WORD_CASE		12
#define ACTION_CUSTOM_INSERT_1		13
#define ACTION_CUSTOM_INSERT_2		14
#define ACTION_CUSTOM_INSERT_3		15
#define ACTION_CHAR_INFO			16
#define ACTION_SCROLLY_UP			17
#define ACTION_SCROLLY_DOWN			18
#define ACTION_FILE_URI				19
#define ACTION_PRINT_VIEWPORT		20
#define ACTION_SAVE_PICTURE			21
#define ACTION_PRINT_HISTORY		22
#define ACTION_PATH_EXPAND			23
#define ACTION_MAX					23

#define LASTARGS_OVERWRITE			1
#define LASTARGS_BREAK_ON_SLASH		2
#define LASTARGS_KEEP_OLD_ARGS		4
#define LASTARGS_SPACE_AROUND		8

#define LASTARGS_DEFAULT_FLAGS		LASTARGS_BREAK_ON_SLASH

#define SCROLLY_DEFAULT_FLAGS		SCROLLY_FLAG_WRAP_AROUND

#define AUTO_CLOSE_BRACKETS			0x0001
#define AUTO_CLOSE_PARENTHESES		0x0002
#define AUTO_CLOSE_BRACES			0x0004
#define AUTO_CLOSE_QUOTES			0x0008
#define AUTO_CLOSE_APOSTROPHES		0x0010
#define AUTO_CLOSE_STRONG_QUOTES	0x0020
#define AUTO_CLOSE_ALL				0x003f
#define AUTO_CLOSE_DEFAULT			0x0000

#define MAX_CUSTOM_STRINGS			3

#define QUOTE						0x22
#define MARKER_CHAR					0xa4

#define SURROGATE_MASK				0xf800
#define SURROGATE_TEST_MASK			0xfc00
#define SURROGATE_VALUE_MASK		0x03ff
#define SURROGATE_OFFSET			0x10000
#define HIGH_SURROGATE				0xd800
#define LOW_SURROGATE				0xdc00
#define MAX_CODEPOINT				0x10ffff



const wchar_t PluginFeatures[] = L"editkeys,*key";



const wchar_t EditKeysHelpText[]	= L"Display and change EditKeys plugin settings. \n\
\n\
EDITKEYS /A:n /B:n /C:n /D /E:n /Kn:key /L /Q /R /S /X:key /Y:n  \r\n\
\n\
\t/A:\tchange the LastArgs flags  \t/A? for more info\n\
\t/B:\tchange the auto-close flags \t/B? for more info\n\
\t/C:\tchange the copy chime \t\t/C? for more info\n\
\t/D\trestore default plugin settings\n\
\t/E:\tchange the error sound \t\t/E? for more info\n\
\t/K:\tdefine or change a hotkey\n\
\t/L\tlist all supported hotkeys\n\
\t/Q\tquietly\n\
\t/R\trestore saved settings from the registry  \n\
\t/S\tsave the current settings to the registry  \n\
\t/X:\tdelete a hotkey\n\
\t/Y:\tchange the Scrolly flags \t/Y? for more info\n\
\n\
EDITKEYS with no options displays the plugin\u2019s current settings.\n";


const wchar_t EditKeys_OptAHelpText[]		= L"EDITKEYS /A:  \u2014 change the LastArgs flags:\n\
\n\
\t\t\u2015\u2015\u2015  Bitmapped  \u2015\u2015\u2015\n\
\t1:\toverwrite text at cursor\n\
\t2:\tbreak at slash\n\
\t4:\tkeep old args\n\
\t8:\tspace around args\n";


const wchar_t EditKeys_OptBHelpText[]		= L"EDITKEYS /B:  \u2014 change the AutoClose flags:\n\
\n\
\t\t\u2015\u2015\u2015  Bitmapped  \u2015\u2015\u2015\n\
\t1\tsquare brackets \t[ ]\n\
\t2\tparentheses  \t\t( )\n\
\t4\tbraces \t\t\t{ }\n\
\t8\tquotation marks \t\" \"\n\
\t16\tapostrophes  \t\t' '\n\
\t32\tstrong quotes \t\t` `\n";


const wchar_t EditKeys_OptCHelpText[]		= L"EDITKEYS /C:  \u2014 change the copy chime sound:\n\
\n\
\t0\tnone\n\
\t1\tchime sound (built into the plugin)\n\
\t2\tthe Windows \u201cAsterisk\u201d sound\n\
\t3\tthe Windows \u201cDefault Beep\u201d sound\n\
\t4\tthe Windows \u201cExclamation\u201d sound\n\
\t5\tthe Windows \u201cCritical Stop\u201d sound\n\
\t6\tthe Windows \u201cQuestion\u201d sound\n";


const wchar_t EditKeys_OptEHelpText[]		= L"EDITKEYS /C:  \u2014 change the error sound:\n\
\n\
\t0\tnone\n\
\t1\tTCC\u2019s internal beep\n\
\t2\tthe Windows \u201cAsterisk\u201d sound\n\
\t3\tthe Windows \u201cDefault Beep\u201d sound\n\
\t4\tthe Windows \u201cExclamation\u201d sound\n\
\t5\tthe Windows \u201cCritical Stop\u201d sound\n\
\t6\tthe Windows \u201cQuestion\u201d sound\n";


const wchar_t EditKeys_OptYHelpText[]		= L"EDITKEYS /Y:  \u2014 Change the Scrolly flags:\n\
\n\
\t\t\u2015\u2015\u2015  Bitmapped  \u2015\u2015\u2015\n\
\t1\twrap around\n\
\t2\tvisible marks\n";



const wchar_t ActionName[ACTION_MAX + 1][24]		= { L"", L"Swap Characters", L"Swap Arguments", L"Insert Last Args", L"Clear Screen",
														L"Math Word", L"Expand Expression", L"Unicode Entry", L"Copy Line", L"Copy Word", L"Cut Word",
														L"Swap Clipboard Word", L"Change Word Case", L"Custom Insert 1", L"Custom Insert 2", L"Custom Insert 3", 
														L"Character Info", L"Scroll Up By Commands", L"Scroll Down By Commands", L"File to URI",
														L"Print Viewport", L"Save Viewport Picture", L"Print History", L"Search Path and Expand", };

const wchar_t ErrorBeepName[ERROR_BEEP_MAX+1][14]	= { L"Off", L"Honk", L"Asterisk", L"Default Beep", L"Exclamation", L"Critical Stop", L"Question" };

const wchar_t CopyBeepName[ERROR_BEEP_MAX+1][14]	= { L"Off", L"Chime", L"Asterisk", L"Default Beep", L"Exclamation", L"Critical Stop", L"Question" };

const wchar_t ColorName[16][15] = { L"black", L"blue", L"green", L"cyan", L"red", L"magenta", L"yellow", L"white",
									L"bright black", L"bright blue", L"bright green", L"bright cyan", L"bright red", L"bright magenta", L"bright yellow", L"bright white" };


const wchar_t PluginRegKey[]			= L"Software\\JPPlugins\\EditKeys";
const wchar_t SaveVersionRegValue[]		= L"Version";
const wchar_t NumHotkeysRegValue[]		= L"Hotkeys";
const wchar_t HotkeyRegValue[]			= L"Key %d";
const wchar_t CustomInsertRegValue[]	= L"Custom Insert String %d";
const wchar_t BeepModeRegValue[]		= L"Error Sound";
const wchar_t CopyBeepRegValue[]		= L"Copy Chime";
const wchar_t LastArgsFlagsRegValue[]	= L"LastArgs Flags";
const wchar_t ScrollyFlagsRegValue[]	= L"Scrolly Flags";
const wchar_t AutoCloseFlagsRegValue[]	= L"Auto Close Flags";

const wchar_t CustomInsertVarName[]		= L"Insert-%u";

const DWORD SaveVersion					= 0x10000;



//		--------------------  Global Variables  --------------------


struct _Hotkey {
	unsigned int nKey;
	unsigned short int Action;
};

_Hotkey Hotkeys[MAX_KEYS];

int NumHotkeys						= 0;
int BeepMode						= ERROR_BEEP_DEFAULT;
int CopyBeep						= COPY_BEEP_DEFAULT;
DWORD LastArgsFlags					= LASTARGS_DEFAULT_FLAGS;
DWORD AutoCloseFlags				= AUTO_CLOSE_DEFAULT;

HANDLE ConsoleHandle				= NULL;

wchar_t *LastArgs					= NULL;
//	wchar_t *LastCmd					= NULL;
BOOL UnicodeEntry					= FALSE;
size_t UnicodeMarker				= 0;

OSVERSIONINFOEX OsVerInfo;

HINSTANCE MyDLLHandle				= NULL;

#ifdef NEED_GDIPLUS
ULONG_PTR GdiplusToken				= NULL;
GdiplusStartupInput StartupInput;
#endif	//	NEED_GDIPLUS

//		------------------------------------------------------------




int DisplayErrorHeader()
{
	wchar_t shortbuf[MAX_ARG_LEN] = L"%_batch";

	Qprintf( ERRHANDLE, L"%s plugin: ", PLUGIN_NAME );

	ExpandVariables( shortbuf, 1 );
	if ( shortbuf[0] == '0' )
		return 0;

	wcscpy_s( shortbuf, MAX_ARG_LEN, L"%_batchname [%_batchline]  " );
	ExpandVariables( shortbuf, 1 );
	Qprintf( ERRHANDLE, shortbuf );

	return 0;
}



int GetErrorText( int err, LPTSTR outString )
{
	if ( !outString )
		return -1;

	if ( err == -1 )
		wcscpy_s( outString, SHORT_BUF_LEN, L"Unknown error" );
	else if ( !FormatMessage( FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS | FORMAT_MESSAGE_MAX_WIDTH_MASK, NULL, err, 0, outString, SHORT_BUF_LEN, NULL ) )
		wsprintf( outString, L"Error %d", err );

	size_t l = wcslen( outString );

	if ( l > 0 )
		while ( iswspace( outString[l - 1] ) ) {
			l--;
			outString[l] = '\0';
		}

	return 0;
}



int GetLastErrorText( LPTSTR outString )
{
	return GetErrorText( GetLastError(), outString );
}



BOOL PipeShell()
{
	wchar_t temp[SHORT_BUF_LEN]			= L"%[_transient]%[_pipe]";


	ExpandVariables( temp, 1 );

	if ( wcscmp( temp, L"00" ) )
		return TRUE;
	else
		return FALSE;
}



LPTSTR _FilenamePart( LPTSTR Filename, LPTSTR outFilename )
{
	if ( !Filename )
		return NULL;
	else if ( Filename[0] == '\0' )
		return NULL;


	BOOL FoundSlash			= FALSE;
	size_t EndOfPathPart	= -1;


	for ( size_t i = 0; Filename[i]; i++ ) {
		if ( ( Filename[i] == '/' ) || ( Filename[i] == '\\' ) ) {
			EndOfPathPart = i;
			FoundSlash = TRUE;
		}
		else if ( ( Filename[i] == ':' ) && !FoundSlash )
			EndOfPathPart = i;
	}

	if ( EndOfPathPart == -1 ) {
		if ( outFilename ) {
			wcscpy_s( outFilename, MAX_FILENAME_LEN, Filename );
			return outFilename;
		}
		else
			return Filename;
	}

	if ( outFilename ) {
		wcscpy_s( outFilename, MAX_FILENAME_LEN, Filename + EndOfPathPart + 1 );
		return outFilename;
	}
	else
		return Filename + EndOfPathPart + 1;
}



int ShowCmdHelp( LPCTSTR HelpText, BOOL ShowVersion, LPCSTR SourceFile, LPCSTR SourceFileVer )
{
	wchar_t File[MAX_FILENAME_LEN]		= L"";
	wchar_t Ver[SHORT_BUF_LEN]			= L"";


	QPuts( (LPTSTR) HelpText );

	if ( ShowVersion ) {
		Printf( L"\r\n\t\t\t\t%s  %u.%u.%u.%u (%s)\r\n", PLUGIN_NAME, VER_MAJOR, VER_MINOR, VER_BUILD, VER_PATCH, PLATFORM );

		if ( SourceFile && SourceFileVer ) {
			MultiByteToWideChar( CP_ACP, MB_PRECOMPOSED, SourceFile, -1, File, MAX_FILENAME_LEN );
			MultiByteToWideChar( CP_ACP, MB_PRECOMPOSED, SourceFileVer, -1, Ver, SHORT_BUF_LEN );
			Printf( L"\t\t\t\t%s  %s\r\n", _FilenamePart( File, NULL), Ver );
		}
	}

	return 0;
}



int ShowCmdHelp( LPCTSTR HelpText, BOOL ShowVersion )
{
	QPuts( (LPTSTR) HelpText );

	if ( ShowVersion )
		Printf( L"\r\n\t\t\t\t%s  %u.%u.%u.%u (%s)\r\n", PLUGIN_NAME, VER_MAJOR, VER_MINOR, VER_BUILD, VER_PATCH, PLATFORM );

	return 0;
}



BOOL IsColonOrEquals( wchar_t ch )
{
	if ( ( ch == ':' ) || ( ch == '=' ) )
		return TRUE;
	else
		return FALSE;
}



int AppendFilename( LPTSTR Path, LPTSTR Filename, size_t BufferSize )
{
	if ( !Path || !Filename )
		return -1;

	
	BOOL NeedSlash		= FALSE;


	for ( size_t i = 0; Path[i]; i++ ) {
		if ( ( Path[i] == '/' ) || ( Path[i] == '\\' ) || ( Path[i] == ':' ) )
			NeedSlash = FALSE;
		else
			NeedSlash = TRUE;
	}

	if ( ( Filename[0] == '/' ) || ( Filename[0] == '\\' ) )
		NeedSlash = FALSE;
	
	if ( NeedSlash )
		wcscat_s( Path, BufferSize, L"\\" );

	wcscat_s( Path, BufferSize, Filename );
	return 0;
}



LPCTSTR _ExtensionPart( LPCTSTR Filename, LPTSTR outExt )
{
	if ( !Filename )
		return NULL;
	else if ( Filename[0] == '\0' )
		return NULL;


	LPCTSTR Ext			= NULL;


	for ( size_t i = 0; Filename[i]; i++ ) {
		if ( ( Filename[i] == '/' ) || ( Filename[i] == '\\' ) || ( Filename[i] == ':' ) )
			Ext = NULL;
		else if ( Filename[i] == '.' )
			Ext = Filename + i;
	}

	if ( !Ext )
		return NULL;

	if ( outExt ) {
		wcscpy_s( outExt, MAX_FILENAME_LEN, Ext );
		return outExt;
	}

	return Ext;
}



//	Strips off any leading or trailing spaces from a string.

int TrimLine( LPTSTR String )
{
	if ( !String )
		return -1;
	else if ( String[0] == '\0' )
		return 0;

	int FirstNonSpace = -1, LastNonSpace = -1;
	wchar_t ch = '\0';

	for ( int i = 0; String[i]; i++ ) {
		ch = String[i];

		if ( !iswspace( ch ) ) {
			LastNonSpace = i;

			if ( FirstNonSpace == -1 )
				FirstNonSpace = i;
		}
	}

	if ( FirstNonSpace == -1 ) {
		String[0] = '\0';
		return 0;
	}

	String[LastNonSpace + 1] = '\0';

	if ( FirstNonSpace ) {
		int i = 0;

		do {
			ch = String[FirstNonSpace + i];
			String[i++] = ch;

		} while ( ch );
	}

	return 0;
}



int StripQuotes( LPTSTR String )
{
	if ( !String )
		return -1;


	int n		= 0;
	size_t j	= 0;


	for ( size_t i = 0; ; i++ ) {
		if ( String[i] == QUOTE ) {
			n++;
			continue;
		}

		String[j++] = String[i];
		if ( String[i] == '\0' )
			break;
	}

	return n;
}



int TrimQuotes( LPTSTR String )
{
	if ( !String )
		return -1;


	size_t len		= wcslen( String );


	if ( len < 2 )
		return 0;

	if ( ( String[0] != QUOTE ) || ( String[len - 1] != QUOTE ) )
		return 0;

	String[len - 1] = '\0';

	for ( unsigned int i = 0; ; i++ ) {
		String[i] = String[i + 1];
		if ( String[i] == '\0' )
			break;
	}

	return 2;
}



int ParseInt( LPTSTR inString, unsigned int *outValue, wchar_t **EndChar )
{
	if ( !inString || !outValue )
		return -1;


	unsigned int value	= 0;
	wchar_t *sep1		= NULL;


	_set_errno( 0 );

	if ( ( inString[0] == '0' ) && ( towupper( inString[1] ) == 'X' ) && ( iswxdigit( inString[2] ) ) ) {
		//  parse hexadecimal

		value = wcstoul( inString + 2, &sep1, 16 );

		if ( errno )
			return -1;
		else if ( !EndChar && *sep1 )
			return -1;

		*outValue = value;

		if ( EndChar )
			*EndChar  = sep1;

		return 0;
	}
	else if ( iswdigit( inString[0] ) ) {
		//  parse unsigned

		value = wcstoul( inString, &sep1, 10 );

		if ( errno )
			return -1;
		else if ( !EndChar &&  *sep1 )
			return -1;

		*outValue = value;

		if ( EndChar )
			*EndChar  = sep1;

		return 0;
	}

	return -1;
}



BOOL IsAFile( LPCTSTR Filename )
{
	if ( !Filename )
		return FALSE;
	else if ( Filename[0] == '\0' )
		return FALSE;


	DWORD attr		= GetFileAttributes( Filename );


	if ( attr == INVALID_FILE_ATTRIBUTES )
		return FALSE;

	if ( attr & ( FILE_ATTRIBUTE_DIRECTORY | FILE_ATTRIBUTE_DEVICE ) )
		return FALSE;

	return TRUE;
}



BOOL GetVersion2( OSVERSIONINFOEXW *VerInfo )
{
	if ( !VerInfo )
		return -1;


	typedef NTSTATUS ( WINAPI _RtlGetVersionProc ) ( OSVERSIONINFOEXW *VerInfo );
	HMODULE NtDllHandle						= GetModuleHandle( L"NtDll.dll" );
	_RtlGetVersionProc *RtlGetVersionProc	= NULL;


	memset( VerInfo, 0x00, sizeof( OSVERSIONINFOEXW ) );
	VerInfo->dwOSVersionInfoSize = sizeof( OSVERSIONINFOEXW );

	if ( NtDllHandle )
		RtlGetVersionProc = (_RtlGetVersionProc*) GetProcAddress( NtDllHandle, "RtlGetVersion" );

	if ( RtlGetVersionProc ) {
		RtlGetVersionProc( VerInfo );
		if ( VerInfo->dwMajorVersion >= 6 )
			return TRUE;
	}

	memset( VerInfo, 0x00, sizeof( OSVERSIONINFOEXW ) );
	VerInfo->dwOSVersionInfoSize = sizeof( OSVERSIONINFOEXW );

	return ( GetVersionEx( (OSVERSIONINFOW*) VerInfo ) != 0 );
}



int OSVerCmp( OSVERSIONINFOEXW *VerInfo, DWORD Major, DWORD Minor, DWORD Build )
{
	if ( !VerInfo )
		return 0;


	if ( VerInfo->dwMajorVersion > Major )
		return 1;
	else if ( VerInfo->dwMajorVersion < Major )
		return -1;
	else if ( VerInfo->dwMinorVersion > Minor )
		return 1;
	else if ( VerInfo->dwMinorVersion < Minor )
		return -1;
	else if ( VerInfo->dwBuildNumber > Build )
		return 1;
	else if ( VerInfo->dwBuildNumber < Build )
		return -1;
	else
		return 0;
}



int ErrorBeep()
{
	switch ( BeepMode ) {

	case 0 :
		break;

	case 2 :
		PlaySound( (LPCWSTR) SND_ALIAS_SYSTEMASTERISK, NULL, SND_ALIAS_ID | SND_ASYNC );
		break;

	case 3 :
		PlaySound( (LPCWSTR) SND_ALIAS_SYSTEMDEFAULT, NULL, SND_ALIAS_ID | SND_ASYNC );
		break;

	case 4 :
		PlaySound( (LPCWSTR) SND_ALIAS_SYSTEMEXCLAMATION, NULL, SND_ALIAS_ID | SND_ASYNC );
		break;

	case 5 :
		PlaySound( (LPCWSTR) SND_ALIAS_SYSTEMHAND, NULL, SND_ALIAS_ID | SND_ASYNC );
		break;

	case 6 :
		PlaySound( (LPCWSTR) SND_ALIAS_SYSTEMQUESTION, NULL, SND_ALIAS_ID | SND_ASYNC );
		break;

	default :
		honk();
	}

	return 0;
}



int CopyChime()
{
	switch( CopyBeep ) {

	case 0 :
		break;

	case 2 :
		PlaySound( (LPCWSTR) SND_ALIAS_SYSTEMASTERISK, NULL, SND_ALIAS_ID | SND_ASYNC );
		break;

	case 3 :
		PlaySound( (LPCWSTR) SND_ALIAS_SYSTEMDEFAULT, NULL, SND_ALIAS_ID | SND_ASYNC );
		break;

	case 4 :
		PlaySound( (LPCWSTR) SND_ALIAS_SYSTEMEXCLAMATION, NULL, SND_ALIAS_ID | SND_ASYNC );
		break;

	case 5 :
		PlaySound( (LPCWSTR) SND_ALIAS_SYSTEMHAND, NULL, SND_ALIAS_ID | SND_ASYNC );
		break;

	case 6 :
		PlaySound( (LPCWSTR) SND_ALIAS_SYSTEMQUESTION, NULL, SND_ALIAS_ID | SND_ASYNC );
		break;

	default :
		PlaySound( (LPCWSTR) IDR_WAVE1, MyDLLHandle, SND_RESOURCE | SND_ASYNC );
	}

	return 0;
}



LPCTSTR LastArgsStr( unsigned int Flags )
{
	static wchar_t temp[128]		= L"";


	temp[0] = '\0';

	if ( Flags & LASTARGS_OVERWRITE )
		wcscpy_s( temp, 128, L"Over" );

	if ( Flags & LASTARGS_BREAK_ON_SLASH ) {
		if ( temp[0] )
			wcscat_s( temp, 128, L", " );

		wcscat_s( temp, 128, L"Slashes" );
	}

	if ( Flags & LASTARGS_KEEP_OLD_ARGS ) {
		if ( temp[0] )
			wcscat_s( temp, 128, L", " );

		wcscat_s( temp, 128, L"KeepOld" );
	}

	if ( Flags & LASTARGS_SPACE_AROUND ) {
		if ( temp[0] )
			wcscat_s( temp, 128, L", " );

		wcscat_s( temp, 128, L"Space" );
	}

	if ( temp[0] == '\0' )
		wcscpy_s( temp, 128, L"none" );

	return temp;
}



LPCTSTR AutoCloseStr( unsigned int Flags )
{
	static wchar_t temp[128]		= L"";


	temp[0] = '\0';

	if ( ( Flags & AUTO_CLOSE_ALL ) == AUTO_CLOSE_ALL ) {
		wcscpy_s( temp, 128, L"All" );
		return temp;
	}

	if ( Flags & AUTO_CLOSE_BRACKETS )
		wcscpy_s( temp, 128, L"Brackets" );

	if ( Flags & AUTO_CLOSE_PARENTHESES ) {
		if ( temp[0] )
			wcscat_s( temp, 128, L", " );

		wcscat_s( temp, 128, L"Parentheses" );
	}

	if ( Flags & AUTO_CLOSE_BRACES ) {
		if ( temp[0] )
			wcscat_s( temp, 128, L", " );

		wcscat_s( temp, 128, L"Braces" );
	}

	if ( Flags & AUTO_CLOSE_QUOTES ) {
		if ( temp[0] )
			wcscat_s( temp, 128, L", " );

		wcscat_s( temp, 128, L"Quotes" );
	}

	if ( Flags & AUTO_CLOSE_APOSTROPHES ) {
		if ( temp[0] )
			wcscat_s( temp, 128, L", " );

		wcscat_s( temp, 128, L"Apostrophes" );
	}

	if ( Flags & AUTO_CLOSE_STRONG_QUOTES ) {
		if ( temp[0] )
			wcscat_s( temp, 128, L", " );

		wcscat_s( temp, 128, L"Strong quotes" );
	}

	if ( temp[0] == '\0' )
		wcscpy_s( temp, 128, L"none" );

	return temp;
}



#include "ParseArgs.cpp"
#include "HotKeys.cpp"
#include "newhelp.cpp"
#include "PluginTempDir.cpp"



BOOL InsertMode()
{
	wchar_t temp[SHORT_BUF_LEN]		= L"%_insert";


	ExpandVariables( temp, 1 );
	return !wcscmp( temp, L"1" );
}



int CmdLineInsert( KEYINFO24 *ki, LPCTSTR String, size_t Position )
{
	wchar_t temp[MAX_CMD_LINE_LEN]		= L"";
	
	size_t Cursor			= ki->pszCurrent - ki->pszLine;


	if ( Position )
		wcsncpy_s( temp, MAX_CMD_LINE_LEN, ki->pszLine, Position );

	wcscat_s( temp, MAX_CMD_LINE_LEN, String );

	if ( Position < wcslen( ki->pszLine ) )
		wcscat_s( temp, MAX_CMD_LINE_LEN, ki->pszLine + Position );

	wcscpy_s( ki->pszLine, MAX_CMD_LINE_LEN, temp );

	if ( Cursor >= Position )
		ki->pszCurrent += wcslen( String );

	//	ClearEditLine( ki->pszLine + Position );
	ki->fRedraw = TRUE;
	return 0;
}



int CmdLineDelete( KEYINFO24 *ki, size_t Start, size_t len )
{
	wchar_t temp[MAX_CMD_LINE_LEN]		= L"";

	size_t Cursor				= ki->pszCurrent - ki->pszLine;


	if ( Start )
		wcsncpy_s( temp, MAX_CMD_LINE_LEN, ki->pszLine, Start );

	if ( ( Start + len ) < wcslen( ki->pszLine ) )
		wcscat_s( temp, MAX_CMD_LINE_LEN, ki->pszLine + Start + len );

	ClearEditLine( ki->pszLine + Start );
	wcscpy_s( ki->pszLine, MAX_CMD_LINE_LEN, temp );

	if ( Cursor >= ( Start + len ) )
		ki->pszCurrent -= len;
	else if ( Cursor > Start )
		ki->pszCurrent = ki->pszLine + Start;

	ki->fRedraw = TRUE;
	return 0;
}



int FindWordAtCursor( KEYINFO24 *ki, size_t *WordStart, size_t *WordLen, LPTSTR outWord, size_t buflen )
{
	if ( !ki )
		return 0;
	else if ( ki->pszLine == NULL )
		return 0;
	else if ( ki->pszLine[0] == '\0' )
		return 0;


	wchar_t ch				= '\0';
	size_t Cursor			= ki->pszCurrent - ki->pszLine;
	size_t Start			= -1;
	size_t End				= -1;
	BOOL quote				= FALSE;
	BOOL InWord				= FALSE;


	for ( size_t i = 0; ki->pszLine[i]; i++ ) {
		ch = ki->pszLine[i];

		if ( ch == QUOTE )
			quote = !quote;

		if ( !iswspace( ch ) && ( i <= Cursor ) ) {
			if ( !InWord )
				Start = i;

			InWord = TRUE;
		}

		if ( ( quote || !iswspace( ch ) ) && InWord )
			End = i;

		if ( iswspace( ch ) && !quote )
			InWord = FALSE;
	}

	if ( ( Start == (size_t) -1 ) || ( ( End + 1 ) < Cursor ) )
		return 0;


	if ( WordStart )
		*WordStart	= Start;

	if ( WordLen )
		*WordLen	= End - Start + 1;

	if ( outWord && ( buflen >= ( End - Start + 2 ) ) )
		wcsncpy_s( outWord, buflen, ki->pszLine + Start, ( End - Start + 1 ) );

	return 1;
}



wchar_t LastChar( LPTSTR String )
{
	if ( !String )
		return '\0';
	else if ( String[0] == '\0' )
		return '\0';


	return String[ wcslen( String ) - 1];
}



int ExpandVariablesQuietly( LPTSTR String )
{
	HANDLE OldStderrHandle	= GetStdHandle( STD_ERROR_HANDLE );
	HANDLE OldStdoutHandle	= GetStdHandle( STD_OUTPUT_HANDLE );
	HANDLE NewStderrHandle	= CreateFile( L"nul", GENERIC_READ | GENERIC_WRITE, FILE_SHARE_READ, NULL, OPEN_ALWAYS, 0, NULL );
	int rv		= 0;


	if ( NewStderrHandle != INVALID_HANDLE_VALUE ) {
		SetStdHandle( STD_ERROR_HANDLE, NewStderrHandle );		//  redirect stderr to nul:
		SetStdHandle( STD_OUTPUT_HANDLE, NewStderrHandle );		//  redirect stdout to nul:
	}
	
	rv = ExpandVariables( String, 1 );							//  and expand the expression

	if ( NewStderrHandle != INVALID_HANDLE_VALUE ) {
		SetStdHandle( STD_ERROR_HANDLE, OldStderrHandle );		//  put stderr back where we found it
		SetStdHandle( STD_OUTPUT_HANDLE , OldStdoutHandle );	//  put stdout back where we found it
		CloseHandle( NewStderrHandle );
	}

	return rv;
}



int GetColorNames( unsigned int Colors, LPTSTR outString )
{
	if ( !outString )
		return -1;

	wcscpy_s( outString, SHORT_BUF_LEN, ColorName[Colors & 0x0f] );
	wcscat_s( outString, SHORT_BUF_LEN, L" on " );
	wcscat_s( outString, SHORT_BUF_LEN, ColorName[ ( Colors & 0xf0 ) >> 4 ] );
	outString[0] = towupper( outString[0] );
	return 0;
}



unsigned int EnvVarValue( LPCTSTR Name, unsigned int Default, unsigned int Min, unsigned int Max )
{
	if ( !Name )
		return -1;
	else if ( Name[0] == '\0' )
		return -1;


	wchar_t temp[SHORT_BUF_LEN]			= L"";

	unsigned int Value					= Default;
	wchar_t *ch							= temp;


	if ( Min > Max ) {
		unsigned int t = Min;
		Min = Max;
		Max = t;
	}

	GetEnvironmentVariable( Name, temp, SHORT_BUF_LEN );
	if ( temp[0] == '\0' )
		return Default;


	while ( iswspace( *ch ) )
		ch++;

	if ( !iswdigit( *ch ) )
		return Default;

	ParseInt( ch, &Value, NULL );

	if ( Value < Min )
		Value = Min;
	else if ( Value > Max )
		Value = Max;

	return Value;
}



unsigned int EnvVarValue( LPCTSTR Name, unsigned int Default )
{
	return EnvVarValue( Name, Default, 0, UINT_MAX );
}



#include "ConsoleMath.cpp"
#include "ClearScreen.cpp"
#include "MathWord.cpp"
#include "ExpandExpression.cpp"
#include "CopyLine.cpp"
#include "SwapArgs.cpp"
#include "SwapClipWord.cpp"
#include "CharInfo.cpp"
#include "FileUriKey.cpp"
#include "Scrolly.cpp"
#include "PrintViewport.cpp"
#include "PrintHistory.cpp"
#include "PathExpand.cpp"



int SwapTwoCharacters( KEYINFO24 *ki )
{
	if ( ki->pszLine == NULL )
		goto Error;


	size_t len				= wcslen( ki->pszLine );

	wchar_t *StartPtr		= ki->pszLine;
	wchar_t *LastPtr		= NULL;
	wchar_t CharA			= '\0';
	wchar_t CharA2			= '\0';
	wchar_t CharB			= '\0';
	wchar_t CharB2			= '\0';
	size_t i				= 0;


	if ( len < 2 )
		goto Error;


	if ( ( ki->pszCurrent[0] & SURROGATE_TEST_MASK ) == HIGH_SURROGATE )
		if ( ( ki->pszCurrent[1] & SURROGATE_TEST_MASK ) == LOW_SURROGATE )
			ki->pszCurrent += 1;


	if ( ( ki->pszCurrent - ki->pszLine ) >= 6 )
		StartPtr = ki->pszCurrent - 6;

	LastPtr = StartPtr;

	for ( wchar_t *ch = LastPtr; *ch; ch++ ) {
		CharA = CharB;
		CharA2 = CharB2;
		StartPtr = LastPtr;

		CharB	= ch[0];
		CharB2	= '\0';

		LastPtr = ch;

		if ( ( ( CharB & SURROGATE_TEST_MASK ) == HIGH_SURROGATE ) && ( ( ch[1] & SURROGATE_TEST_MASK ) == LOW_SURROGATE ) )
			CharB2 = ch[1], ch++;

		if ( CharA && CharB && ( ch >= ki->pszCurrent ) )
			break;
	}


	if ( CharA && CharB ) {
		StartPtr[i++] = CharB;
		if ( CharB2 )
			StartPtr[i++] = CharB2;

		StartPtr[i++] = CharA;
		if ( CharA2 )
			StartPtr[i++] = CharA2;
	}
	else
		goto Error;


	ClearEditLine( StartPtr );
	ki->fRedraw = 1;
	ki->nKey = 0;
	return 0;

Error :

	ErrorBeep();
	ki->nKey = 0;
	return 0;
}



LPTSTR StartArgs( LPTSTR Line )
{
	if ( !Line )
		return NULL;
	else if ( Line[0] == '\0' )
		return NULL;


	wchar_t *ch			= Line;
	BOOL Quote			= FALSE;
	BOOL SlashAtStart	= FALSE;


	while ( iswspace( *ch ) )
		ch++;

	if ( *ch == '/' )
		SlashAtStart = TRUE;


	do {
		if ( *ch == '\0' )
			return NULL;
		
		if ( *ch == QUOTE )
			Quote = !Quote;

		if ( iswspace( *ch ) && !Quote )
			break;

		if ( ( LastArgsFlags & LASTARGS_BREAK_ON_SLASH ) && !Quote && !SlashAtStart && ( *ch == '/' ) )
			break;

		ch++;

	} while ( TRUE );


	while ( iswspace( *ch ) )
		ch++;

	if ( *ch == '\0' )
		return NULL;

	return ch;
}



int SaveCmdArgs( KEYINFO24 *KeyInfo )
{
	if ( KeyInfo->pszLine == NULL )
		return 0;


	wchar_t *Args			= StartArgs( KeyInfo->pszLine );
	size_t len				= 0;


	if ( !Args ) {
		if ( ( LastArgsFlags & LASTARGS_KEEP_OLD_ARGS ) == 0 ) {
			free( LastArgs );
			LastArgs = NULL;
		}

		return 0;
	}


	len = wcslen( Args ) + 1;

	free( LastArgs );
	LastArgs = (wchar_t*) malloc( len * sizeof( wchar_t ) );

	if ( !LastArgs ) {
		ErrorBeep();
		return 0;
	}

	wcscpy_s( LastArgs, len, Args );
	return 0;
}



int InsertLastArgs( KEYINFO24 *KeyInfo )
{
	wchar_t NewLine[MAX_CMD_LINE_LEN]		= L"";

	size_t CursorPos		= ( KeyInfo->pszCurrent - KeyInfo->pszLine );
	unsigned int CursorAdj	= 0;
	
	
	KeyInfo->nKey = 0;
	if ( NewKeyHook )
		KeyInfo->pszKey[0] = '\0';


	if ( !LastArgs ) {
		ErrorBeep();
		return 0;
	}


	if ( CursorPos ) {
		wcsncpy_s( NewLine, MAX_CMD_LINE_LEN, KeyInfo->pszLine, CursorPos );

		if ( ( LastArgsFlags & LASTARGS_SPACE_AROUND ) && !iswspace( LastChar( NewLine ) ) ) {
			wcscat_s( NewLine, MAX_CMD_LINE_LEN, L" " );
			CursorAdj++;
		}
	}

	wcscat_s( NewLine, MAX_CMD_LINE_LEN, LastArgs );


	if ( LastArgsFlags & LASTARGS_OVERWRITE ) {
		ClearEditLine( KeyInfo->pszCurrent );
		wcscpy_s( KeyInfo->pszLine, MAX_CMD_LINE_LEN, NewLine );
		KeyInfo->pszCurrent = KeyInfo->pszLine + wcslen( KeyInfo->pszLine );
		KeyInfo->fRedraw = 1;
	}
	else {
		if ( KeyInfo->pszCurrent[0] ) {
			if ( ( LastArgsFlags & LASTARGS_SPACE_AROUND ) && !iswspace( LastChar( NewLine ) ) && !iswspace( KeyInfo->pszCurrent[0] ) ) {
				wcscat_s( NewLine, MAX_CMD_LINE_LEN, L" " );
				CursorAdj++;
			}

			wcscat_s( NewLine, MAX_CMD_LINE_LEN, KeyInfo->pszCurrent );
		}

		ClearEditLine( KeyInfo->pszCurrent );
		wcscpy_s( KeyInfo->pszLine, MAX_CMD_LINE_LEN, NewLine );
		KeyInfo->pszCurrent += ( wcslen( LastArgs ) + CursorAdj );
		KeyInfo->fRedraw = 1;
	}

	return 0;
}



/*
int InsertLastArgsLastWord( KEYINFO24 *KeyInfo )
{
	wchar_t NewLine[MAX_CMD_LINE_LEN]		= L"";

	size_t CursorPos		= KeyInfo->pszCurrent - KeyInfo->pszLine;
	unsigned int CursorAdj	= 0;

	wchar_t *LastWord		= NULL;
	size_t EndLastWord		= 0;
	BOOL InWord				= FALSE;
	BOOL quote				= FALSE;


	KeyInfo->nKey = 0;
	if ( NewKeyHook )
		KeyInfo->pszKey[0] = '\0';


	if ( !LastArgs ) {
		ErrorBeep();
		return 0;
	}


	for ( wchar_t *ch = LastArgs; *ch; ch++ ) {

		if ( *ch == QUOTE )
			quote = !quote;

		if ( !iswspace( *ch ) ) {
			if ( !InWord )
				LastWord = ch;

			InWord = TRUE;
			EndLastWord = 0;
		}
		else if ( !quote && iswspace( *ch ) ) {
			if ( InWord )
				EndLastWord = ch - LastWord;

			InWord = FALSE;
		}
	}

	if ( !LastWord ) {
		ErrorBeep();
		return 0;
	}


	if ( CursorPos ) {
		wcsncpy_s( NewLine, MAX_CMD_LINE_LEN, KeyInfo->pszLine, CursorPos );

		if ( ( LastArgsFlags & LASTARGS_SPACE_AROUND ) && !iswspace( LastChar( NewLine ) ) ) {
			wcscat_s( NewLine, MAX_CMD_LINE_LEN, L" " );
			CursorAdj++;
		}
	}

	wcscat_s( NewLine, MAX_CMD_LINE_LEN, LastWord );
	if ( EndLastWord )
		NewLine[CursorPos + CursorAdj + EndLastWord] = '\0';

	CursorAdj = (unsigned int) wcslen( NewLine + CursorPos );

	if ( LastArgsFlags & LASTARGS_OVERWRITE ) {
		ClearEditLine( KeyInfo->pszCurrent );
		wcscpy_s( KeyInfo->pszLine, MAX_CMD_LINE_LEN, NewLine );
		KeyInfo->fRedraw = 1;
	}
	else {
		if ( KeyInfo->pszCurrent[0] ) {
			if ( ( LastArgsFlags & LASTARGS_SPACE_AROUND ) && !iswspace( LastChar( NewLine ) ) ) {
				wcscat_s( NewLine, MAX_CMD_LINE_LEN, L" " );
				CursorAdj++;
			}

			wcscat_s( NewLine, MAX_CMD_LINE_LEN, KeyInfo->pszCurrent );
		}

		ClearEditLine( KeyInfo->pszCurrent );
		wcscpy_s( KeyInfo->pszLine, MAX_CMD_LINE_LEN, NewLine );
		KeyInfo->pszCurrent += CursorAdj;
		KeyInfo->fRedraw = 1;
	}

	return 0;
}
*/



int ChangeWordCase( KEYINFO24 *ki )
{
	size_t WordStart		= 0;
	size_t WordLen			= 0;

	wchar_t *ch				= NULL;
	unsigned int nUpper		= 0;
	unsigned int nLower		= 0;
	BOOL FirstLetter		= TRUE;


	if ( FindWordAtCursor( ki, &WordStart, &WordLen, NULL, 0 ) && WordLen ) {

		//	examine the word :

		for ( size_t i = WordStart; i < ( WordStart + WordLen ); i++ ) {
			if ( IsCharAlphaW( ki->pszLine[i] ) ) {
				if ( !ch )
					ch = ki->pszLine + i;						//	find the first letter

				if ( IsCharUpperW( ki->pszLine[i] ) )
					nUpper++;									//	and count uppercase
				else if ( IsCharLowerW( ki->pszLine[i] ) )
					nLower++;									//	and lowercase letters
			}
		}


		if ( ch ) {
			if ( nUpper && nLower ) {							//	mixed case:  change word to uppercase
				for ( wchar_t *c = ki->pszLine + WordStart; c < ( ki->pszLine + WordStart + WordLen ); c++ )
					*c = (wchar_t) CharUpperW( (LPWSTR) *c );
			}
			else if ( nLower ) {								//	change word to initial caps
				for ( wchar_t *c = ki->pszLine + WordStart; c < ( ki->pszLine + WordStart + WordLen); c++ ) {
					if ( iswspace( *c ) || ( *c == '-' ) )
						FirstLetter = TRUE;
					else if ( IsCharAlphaW( *c ) ) {
						if ( FirstLetter )
							*c = (wchar_t) CharUpperW( (LPWSTR) *c );
						else
							*c = (wchar_t) CharLowerW( (LPWSTR) *c );

						FirstLetter = FALSE;
					}
				}
			}
			else if ( nUpper ) {								//	change word to lowercase
				for ( wchar_t *c = ki->pszLine + WordStart; c < ( ki->pszLine + WordStart + WordLen ); c++ )
					*c = (wchar_t) CharLowerW( (LPWSTR) *c );
			}

			ClearEditLine( ki->pszLine + WordStart );
			ki->pszCurrent = ki->pszLine + WordStart + WordLen;
			ki->fRedraw = 1;
		}
	}


	if ( !ch )
		ErrorBeep();


	ki->nKey = 0;
	if ( NewKeyHook )
		ki->pszKey[0] = '\0';

	return 0;
}



/*
int SaveLastCommand( KEYINFO24 *ki )
{
	if ( ki->pszLine == NULL )
		return 0;


	wchar_t *Start			= NULL;
	wchar_t *End			= NULL;
	BOOL quote				= FALSE;
	size_t len				= 0;


	for ( wchar_t *ch = ki->pszLine; *ch; ch++ ) {

		if ( *ch == QUOTE )
			quote = !quote;

		if ( !quote && iswspace( *ch ) ) {
			if ( Start ) {
				End = ch;
				break;
			}
		}
		else {
			if ( !Start )
				Start = ch;
		}
	}


	if ( LastCmd ) {
		free( LastCmd );
		LastCmd = NULL;
	}


	if ( !Start )
		return 0;

	if ( End )
		len = ( End - Start ) + 1;
	else
		len = wcslen( Start ) + 1;


	LastCmd = (wchar_t*) malloc( len * sizeof( wchar_t ) );

	if ( LastCmd )
		wcsncpy_s( LastCmd, len, Start, len - 1 );

	return 0;
}
*/



int EndUnicodeEntry( KEYINFO24 *ki, BOOL SwallowKey )
{
	if ( !UnicodeEntry )
		return 0;


	wchar_t NewChar[4]	= L"";

	size_t Cursor		= ki->pszCurrent - ki->pszLine;

	unsigned int Value	= 0;
	wchar_t *sep1		= NULL;


	UnicodeEntry = FALSE;


	if ( UnicodeMarker >= Cursor )
		goto Error;
	else if ( ki->pszLine[UnicodeMarker] != MARKER_CHAR )
		goto Error;


	if ( SwallowKey ) {
		ki->nKey = 0;

		if ( NewKeyHook )
			ki->pszKey[0] = '\0';
	}


	if ( !iswxdigit( ki->pszLine[UnicodeMarker + 1] ) ) {
		CmdLineDelete( ki, UnicodeMarker, 1 );
		goto Error;
	}

	Value = wcstoul( ki->pszLine + UnicodeMarker + 1, &sep1, 16 );
	
	if ( !Value || ( Value > MAX_CODEPOINT ) ) {
		CmdLineDelete( ki, UnicodeMarker, sep1 - ( ki->pszLine + UnicodeMarker ) );
		goto Error;
	}


	if ( Value < 0x10000 ) {
		NewChar[0]	= Value;
		NewChar[1]	= '\0';
	}
	else {
		Value -= 0x10000;
		NewChar[0]	= ( Value >> 10 ) | HIGH_SURROGATE;
		NewChar[1]	= ( Value & 0x03ff ) | LOW_SURROGATE;
		NewChar[2]	= '\0';
	}


	CmdLineDelete( ki, UnicodeMarker, sep1 - ( ki->pszLine + UnicodeMarker ) );
	CmdLineInsert( ki, NewChar, UnicodeMarker );
	return 1;


Error :

	ErrorBeep();
	return 0;
}



int BeginUnicodeEntry( KEYINFO24 *ki )
{
	if ( UnicodeEntry )
		EndUnicodeEntry( ki, FALSE );


	UnicodeMarker = ki->pszCurrent - ki->pszLine;
	CmdLineInsert( ki, L"\u00a4", UnicodeMarker );
	UnicodeEntry = TRUE;

	ki->nKey = 0;
	if ( NewKeyHook )
		ki->pszKey[0] = '\0';

	return 0;
}



int CustomInsert( KEYINFO24 *ki, unsigned int n )
{
	if ( n >= MAX_CUSTOM_STRINGS )
		return -1;


	wchar_t VarName[SHORT_BUF_LEN]		= L"";
	wchar_t VarBuf[MAX_ARG_LEN]			= L"";


	wsprintf( VarName, CustomInsertVarName, n + 1 );
	GetEnvironmentVariable( VarName, VarBuf, MAX_ARG_LEN );

	if ( VarBuf[0] == '\0' )
		wcscpy_s( VarBuf, MAX_ARG_LEN, L"%_isodate" );


	if ( wcschr( VarBuf, '%' ) ) {
		int rv = ExpandVariablesQuietly( VarBuf );

		if ( rv || ( VarBuf[0] == '\0' ) ) {
			ki->nKey = 0;
			if ( NewKeyHook )
				ki->pszKey[0] = '\0';

			ErrorBeep();
			return 0;
		}
	}


	CmdLineInsert( ki, VarBuf, ki->pszCurrent - ki->pszLine );

	ki->nKey = 0;
	if ( NewKeyHook )
		ki->pszKey[0] = '\0';

	return 0;
}



int AutoClose( KEYINFO24 *ki )
{
	if ( !InsertMode() )
		return 0;


	wchar_t PrevChar			= '\0';
	wchar_t CloseChar			= '\0';
	wchar_t *ch					= ki->pszCurrent;


	if ( ki->pszCurrent > ki->pszLine )
		PrevChar = ki->pszLine[ki->pszCurrent - ki->pszLine - 1];

	if ( ( ki->nKey == '[' ) && ( AutoCloseFlags & AUTO_CLOSE_BRACKETS ) )
		CloseChar = ']';
	else if ( ( ki->nKey == '(' ) && ( AutoCloseFlags & AUTO_CLOSE_PARENTHESES ) )
		CloseChar = ')';
	else if ( ( ki->nKey == '{' ) && ( AutoCloseFlags & AUTO_CLOSE_BRACES ) )
		CloseChar = '}';
	else if ( ( ki->nKey == '\"' ) && ( AutoCloseFlags & AUTO_CLOSE_QUOTES ) ) {
		if ( !PrevChar || iswspace( PrevChar ) || ( PrevChar == ':' ) || ( PrevChar == '=' ) || ( PrevChar == '[' ) )
			CloseChar = '\"';
	}
	else if ( ( ki->nKey == '\'' ) && ( AutoCloseFlags & AUTO_CLOSE_APOSTROPHES ) ) {
		if ( !PrevChar || iswspace( PrevChar ) || iswpunct( PrevChar ) )
			CloseChar = '\'';
	}
	else if ( ( ki->nKey == '`' ) && ( AutoCloseFlags & AUTO_CLOSE_STRONG_QUOTES ) ) {
		if ( !PrevChar || iswspace( PrevChar ) || ( PrevChar == ':' ) || ( PrevChar == '=' ) || ( PrevChar == '[' ) )
			CloseChar = '`';
	}
	
	if ( !CloseChar )
		return 0;
	else if ( wcslen( ki->pszLine ) > ( MAX_CMD_LINE_LEN - 3 ) )
		return 0;


	for ( wchar_t *c = ch + wcslen( ch ); c >= ch; c-- )
		c[1] = c[0];

	ch[0]	= CloseChar;
	return 0;
}



int RestorePluginDefaults()
{
	NumHotkeys			= 0;
	memset( Hotkeys, 0x00, MAX_KEYS * sizeof( _Hotkey ) );

	Hotkeys[0].nKey		= 0x10027;		//	Alt-;
	Hotkeys[0].Action	= ACTION_SWAP_CHARS;

	Hotkeys[1].nKey		= 0x10028;		//	Alt-'
	Hotkeys[1].Action	= ACTION_SWAP_ARGS;

	Hotkeys[2].nKey		= 0x1003d;		//	F3
	Hotkeys[2].Action	= ACTION_INSERT_LAST_ARGS;

	Hotkeys[3].nKey		= 0x10026;		//	Alt-L
	Hotkeys[3].Action	= ACTION_CLEAR_SCREEN;

	Hotkeys[4].nKey		= 0x10083;		//	Alt-=
	Hotkeys[4].Action	= ACTION_MATH_WORD;

	Hotkeys[5].nKey		= 0x20018;		//	Ctrl-Shift-X
	Hotkeys[5].Action	= ACTION_EXPAND_EXPR;

	Hotkeys[6].nKey		= 0x10016;		//	Alt-U
	Hotkeys[6].Action	= ACTION_UNICODE_ENTRY;

	Hotkeys[7].nKey		= 0x1002e;		//	Alt-C
	Hotkeys[7].Action	= ACTION_COPY_LINE;

	Hotkeys[8].nKey		= 0x10011;		//	Alt-W
	Hotkeys[8].Action	= ACTION_COPY_WORD;

	Hotkeys[9].nKey		= 0x50011;		//	Ctrl-Alt-W
	Hotkeys[9].Action	= ACTION_CUT_WORD;

	Hotkeys[10].nKey	= 0x5002c;		//	Ctrl-Alt-Z
	Hotkeys[10].Action	= ACTION_SWAP_CLIP_WORD;

	Hotkeys[11].nKey	= 0x15;			//	Ctrl-U
	Hotkeys[11].Action	= ACTION_CHANGE_WORD_CASE;

	Hotkeys[12].nKey	= 0x10017;		//	Alt-I
	Hotkeys[12].Action	= ACTION_CUSTOM_INSERT_1;

	Hotkeys[13].nKey	= 0x01;			//	Ctrl-A
	Hotkeys[13].Action	= ACTION_CHAR_INFO;

	Hotkeys[14].nKey	= 0x5008d;		//	Ctrl-Alt-Up
	Hotkeys[14].Action	= ACTION_SCROLLY_UP;

	Hotkeys[15].nKey	= 0x50091;		//	Ctrl-Alt-Down
	Hotkeys[15].Action	= ACTION_SCROLLY_DOWN;

	Hotkeys[16].nKey	= 0x50021;		//	Ctrl-Alt-F
	Hotkeys[16].Action	= ACTION_FILE_URI;

	Hotkeys[17].nKey	= 0x20010;		//	Ctrl-Shift-P
	Hotkeys[17].Action	= ACTION_PRINT_VIEWPORT;

	Hotkeys[18].nKey	= 0x20011;		//	Ctrl-Shift-Q
	Hotkeys[18].Action	= ACTION_SAVE_PICTURE;

	Hotkeys[19].nKey	= 0x10;			//	Ctrl-P
	Hotkeys[19].Action	= ACTION_PRINT_HISTORY;

	Hotkeys[20].nKey	= 0x50012;		//	Ctrl-Alt-E
	Hotkeys[20].Action	= ACTION_PATH_EXPAND;

	NumHotkeys			= 21;
	BeepMode			= ERROR_BEEP_DEFAULT;
	CopyBeep			= COPY_BEEP_DEFAULT;
	LastArgsFlags		= LASTARGS_DEFAULT_FLAGS;
	ScrollyFlags		= SCROLLY_DEFAULT_FLAGS;
	AutoCloseFlags		= AUTO_CLOSE_DEFAULT;

	return 0;
}



DWORD RegGetDWord( HKEY KeyHandle, LPCTSTR ValueName, DWORD Min, DWORD Max, DWORD Default, int *Error )
{
	DWORD value			= -1;
	DWORD bufsize		= sizeof( DWORD );
	DWORD valtype		= 0;


	RegQueryValueEx( KeyHandle, ValueName, NULL, &valtype, (BYTE*) &value, &bufsize );

	if ( ( valtype != REG_DWORD ) || ( value < Min ) || ( value > Max ) ) {
		value = Default;

		if ( Error )
			*Error += 1;
	}

	return value;
}



int RegRestoreEnvVar( HKEY KeyHandle, LPCTSTR VarName, BOOL DeleteIfEmpty )
{
	if ( !KeyHandle || !VarName )
		return -1;
	else if ( VarName[0] == '\0' )
		return -1;


	wchar_t temp[SHORT_BUF_LEN]			= L"";

	DWORD ValType		= 0;
	DWORD ValSize		= SHORT_BUF_LEN * sizeof( wchar_t );
	int rv				= 0;


	if ( RegQueryValueEx( KeyHandle, VarName, NULL, &ValType, (BYTE*) temp, &ValSize ) )
		temp[0] = '\0';
	else if ( ValType != REG_SZ )
		temp[0] = '\0';


	if ( temp[0] ) {
		if ( !SetEnvironmentVariable( VarName, temp ) )
			rv = 1;
	}
	else if ( DeleteIfEmpty )
		SetEnvironmentVariable( VarName, NULL );

	return rv;
}



int RegSaveEnvVar( HKEY KeyHandle, LPCTSTR VarName )
{
	if ( !KeyHandle || !VarName )
		return -1;
	else if ( VarName[0] == '\0' )
		return -1;


	wchar_t temp[SHORT_BUF_LEN]			= L"";

	int rv			= 0;


	GetEnvironmentVariable( VarName, temp, SHORT_BUF_LEN );

	if ( temp[0] ) {
		if ( RegSetValueEx( KeyHandle, VarName, 0, REG_SZ, (BYTE*) temp, (DWORD) ( wcslen( temp ) + 1 ) * sizeof( wchar_t ) ) )
			rv = 1;
	}
	else
		RegDeleteValue( KeyHandle, VarName );

	return rv;
}



int RestoreOptions( HKEY RootKey )
{
	wchar_t ValName[SHORT_BUF_LEN]								= L"";
	wchar_t NewCustomString[MAX_CUSTOM_STRINGS][MAX_ARG_LEN]	= { L"", L"", L"" };
	BOOL GotCustomString[MAX_CUSTOM_STRINGS]					= { FALSE, FALSE, FALSE };

	HKEY KeyHandle					= NULL;
	DWORD newnumhotkeys				= -1;
	DWORD newbeepmode				= -1;
	DWORD newcopybeep				= -1;
	DWORD newlastargsflags			= -1;
	DWORD newscrollyflags			= -1;
	DWORD newautocloseflags			= -1;

	_Hotkey newhotkeys[MAX_KEYS];

	DWORD savever					= 0;

	DWORD ValType					= 0;
	DWORD ValSize					= 0;


	int rv							= 0;


	if ( RegOpenKeyEx( RootKey, PluginRegKey, 0, KEY_READ, &KeyHandle ) != ERROR_SUCCESS )
		return 1;


	savever				= RegGetDWord( KeyHandle, SaveVersionRegValue, 0, UINT_MAX, -1, &rv );
	
	if ( rv || ( ( savever >> 16 ) != 1 ) || ( ( savever & 0x0000ffff ) != 0 ) ) {
		RegCloseKey( KeyHandle );
		return -10;
	}


	newnumhotkeys		= RegGetDWord( KeyHandle, NumHotkeysRegValue, 0, MAX_KEYS, -1, &rv );
	newbeepmode			= RegGetDWord( KeyHandle, BeepModeRegValue, 0, ERROR_BEEP_MAX, ERROR_BEEP_DEFAULT, &rv );
	newcopybeep			= RegGetDWord( KeyHandle, CopyBeepRegValue, 0, COPY_BEEP_MAX, COPY_BEEP_DEFAULT, &rv );
	newlastargsflags	= RegGetDWord( KeyHandle, LastArgsFlagsRegValue, 0, UINT_MAX, LASTARGS_DEFAULT_FLAGS, &rv );
	newscrollyflags		= RegGetDWord( KeyHandle, ScrollyFlagsRegValue, 0, UINT_MAX, SCROLLY_DEFAULT_FLAGS, &rv );
	newautocloseflags	= RegGetDWord( KeyHandle, AutoCloseFlagsRegValue, 0, UINT_MAX, AUTO_CLOSE_DEFAULT, &rv );


	if ( !rv && newnumhotkeys ) {
		for ( int i = 0; i < (int) newnumhotkeys; i++ ) {
			wsprintf( ValName, HotkeyRegValue, i );

			ValType = 0;
			ValSize = sizeof( _Hotkey );

			if ( RegQueryValueEx( KeyHandle, ValName, NULL, &ValType, (BYTE*) ( newhotkeys + i ), &ValSize ) != ERROR_SUCCESS )
				rv++;
			else if ( ( ValType != REG_BINARY ) || ( ValSize != sizeof( _Hotkey ) ) )
				rv++;
		}
	}

	if ( !rv ) {
		for ( int i = 0; i < MAX_CUSTOM_STRINGS; i++ ) {
			wsprintf( ValName, CustomInsertRegValue, i + 1 );

			ValType = 0;
			ValSize = MAX_ARG_LEN * sizeof( wchar_t );

			if ( RegQueryValueEx( KeyHandle, ValName, NULL, &ValType, (BYTE*) NewCustomString[i], &ValSize ) == ERROR_SUCCESS ) {
				GotCustomString[i] = TRUE;

				if ( ValType != REG_SZ ) {
					NewCustomString[i][0] = '\0';
					GotCustomString[i] = FALSE;
				}
			}
			else {
				NewCustomString[i][0] = '\0';
				GotCustomString[i] = FALSE;
			}
		}

#ifdef PRINT_VIEWPORT_VER

		RegRestoreEnvVar( KeyHandle, PVP_SavePicDirVarName, FALSE );
		RegRestoreEnvVar( KeyHandle, PVP_SavePicExtVarName, FALSE );
		RegRestoreEnvVar( KeyHandle, PVP_SavePicFlagsVarName, FALSE );

#endif	//	PRINT_VIEWPORT_VER

#ifdef PRINT_HISTORY_VER

		RegRestoreEnvVar( KeyHandle, PH_MaxLinesVarName, FALSE );
		RegRestoreEnvVar( KeyHandle, PH_FlagsVarName, FALSE );

#endif	//	PRINT_HISTORY_VER

#ifdef PATH_EXPAND_VER

		RegRestoreEnvVar( KeyHandle, PathExp_FlagsVarName, FALSE );

#endif	//	PATH_EXPAND_VER

	}


	if ( !rv ) {
		NumHotkeys	= 0;

		memset( Hotkeys, 0x00, MAX_KEYS * sizeof( _Hotkey ) );

		if ( newnumhotkeys )
			memcpy( Hotkeys, newhotkeys, newnumhotkeys * sizeof( _Hotkey ) );

		NumHotkeys		= newnumhotkeys;
		BeepMode		= newbeepmode;
		CopyBeep		= newcopybeep;
		LastArgsFlags	= newlastargsflags;
		ScrollyFlags	= newscrollyflags;
		AutoCloseFlags	= newautocloseflags;

		for ( int i = 0; i < MAX_CUSTOM_STRINGS; i++ ) {
			if ( GotCustomString[i] ) {
				wsprintf( ValName, CustomInsertVarName, i + 1 );
				SetEnvironmentVariable( ValName, NewCustomString[i][0] ? NewCustomString[i] : NULL );
			}
		}
	}

	RegCloseKey( KeyHandle );
	return rv;
}



HKEY RestoreOptions()
{
	if ( RestoreOptions( HKEY_CURRENT_USER ) == 0 )
		return HKEY_CURRENT_USER;
	
	if ( RestoreOptions( HKEY_LOCAL_MACHINE ) == 0 )
		return HKEY_LOCAL_MACHINE;

	return NULL;
}



int SaveOptions( HKEY RootKey )
{
	wchar_t ValName[SHORT_BUF_LEN]	= L"";
	wchar_t VarName[SHORT_BUF_LEN]	= L"";
	wchar_t VarValue[MAX_ARG_LEN]	= L"";
	HKEY KeyHandle					= NULL;

	int t							= 0;
	int rv							= 0;


	if ( ( t = RegCreateKeyEx( RootKey, PluginRegKey, 0, NULL, 0, KEY_READ | KEY_WRITE, NULL, &KeyHandle, NULL ) ) == ERROR_SUCCESS ) {

		if ( t = RegSetValueEx( KeyHandle, SaveVersionRegValue, 0, REG_DWORD, (BYTE*) &SaveVersion, sizeof( DWORD ) ) )
			rv = t;

		if ( t = RegSetValueEx( KeyHandle, NumHotkeysRegValue, 0, REG_DWORD, (BYTE*) &NumHotkeys, sizeof( DWORD ) ) )
			rv = t;

		if ( t = RegSetValueEx( KeyHandle, BeepModeRegValue, 0, REG_DWORD, (BYTE*) &BeepMode, sizeof( DWORD ) ) )
			rv = t;

		if ( t = RegSetValueEx( KeyHandle, CopyBeepRegValue, 0, REG_DWORD, (BYTE*) &CopyBeep, sizeof( DWORD ) ) )
			rv = t;

		if ( t = RegSetValueEx( KeyHandle, LastArgsFlagsRegValue, 0, REG_DWORD, (BYTE*) &LastArgsFlags, sizeof( DWORD ) ) )
			rv = t;

		if ( t = RegSetValueEx( KeyHandle, ScrollyFlagsRegValue, 0, REG_DWORD, (BYTE*) &ScrollyFlags, sizeof( DWORD ) ) )
			rv = t;

		if ( t = RegSetValueEx( KeyHandle, AutoCloseFlagsRegValue, 0, REG_DWORD, (BYTE*) &AutoCloseFlags, sizeof( DWORD ) ) )
			rv = t;


		for ( int i = 0; i < MAX_KEYS; i++ ) {
			wsprintf( ValName, HotkeyRegValue, i );

			if ( i < NumHotkeys ) {
				if ( t = RegSetValueEx( KeyHandle, ValName, 0, REG_BINARY, (BYTE*) ( Hotkeys + i ), sizeof( _Hotkey ) ) )
					rv = t;
			}
			else
				RegDeleteValue( KeyHandle, ValName );
		}

		for ( int i = 0; i < MAX_CUSTOM_STRINGS; i++ ) {
			wsprintf( ValName, CustomInsertRegValue, i + 1 );
			wsprintf( VarName, CustomInsertVarName, i + 1 );

			VarValue[0] = '\0';
			GetEnvironmentVariable( VarName, VarValue, MAX_ARG_LEN );

			if ( VarValue[0] ) {
				if ( t = RegSetValueEx( KeyHandle, ValName, 0, REG_SZ, (BYTE*) VarValue, (DWORD) ( ( wcslen( VarValue ) + 1 ) * sizeof( wchar_t ) ) ) )
					rv = t;
			}
			else
				RegDeleteValue( KeyHandle, ValName );
		}

#ifdef PRINT_VIEWPORT_VER

		RegSaveEnvVar( KeyHandle, PVP_SavePicDirVarName );
		RegSaveEnvVar( KeyHandle, PVP_SavePicExtVarName );
		RegSaveEnvVar( KeyHandle, PVP_SavePicFlagsVarName );

#endif	//	PRINT_VIEWPORT_VER

#ifdef PRINT_HISTORY_VER

		RegSaveEnvVar( KeyHandle, PH_MaxLinesVarName );
		RegSaveEnvVar( KeyHandle, PH_FlagsVarName );

#endif	//	PRINT_HISTORY_VER

#ifdef PATH_EXPAND_VER

		RegSaveEnvVar( KeyHandle, PathExp_FlagsVarName );

#endif	//	PATH_EXPAND_VER

		RegCloseKey( KeyHandle );
	}
	else
		rv = t;

	return rv;
}



LPCTSTR RootKeyName( HKEY RootKey )
{
	static const wchar_t RootName[2][5]		= { L"HKCU", L"HKLM" };


	if ( RootKey == HKEY_CURRENT_USER )
		return RootName[0];
	else if ( RootKey == HKEY_LOCAL_MACHINE )
		return RootName[1];
	else
		return NULL;
}



LPCTSTR Dots( int n )
{
	static wchar_t temp[MAX_DOTS]		= L"";

	static BOOL init			= FALSE;


	if ( !init ) {
		for ( int i = 0; i < MAX_DOTS; i += 2 ) {
			temp[i] = '.';
			temp[i + 1] = 0x20;
		}

		temp[MAX_DOTS - 1] = '\0';
		init = TRUE;
	}


	if ( ( n < 1 ) || ( n >= MAX_DOTS ) )
		return temp + MAX_DOTS - 1;

	return temp + MAX_DOTS - n - 1;
}



int ShowHotkeyAssignments()
{
	wchar_t keyname[SHORT_BUF_LEN]		= L"";
	wchar_t shortbuf[SHORT_BUF_LEN]		= L"";


	for ( int a = 1; a <= ACTION_MAX; a++ ) {
		if ( ActionName[a][0] == '\0' )
			continue;

		shortbuf[0] = '\0';

		for ( int i = 0; i < NumHotkeys; i++ ) {
			if ( Hotkeys[i].Action == a ) {
				keyname[0] = '\0';
				GetKeyName( Hotkeys[i].nKey, keyname, TRUE );

				if ( keyname[0] ) {
					if ( shortbuf[0] )
						wcscat_s( shortbuf, SHORT_BUF_LEN, L" \u00b7 " );

					wcscat_s( shortbuf, SHORT_BUF_LEN, keyname );
				}
			}
		}

		Printf( L"  %s %s /K%-2d : %s\r\n", ActionName[a], Dots( 29 - (int) wcslen( ActionName[a] ) ), a, shortbuf );
	}

	return 0;
}



int DeleteHotkey( int nKey, _Hotkey *Hotkeys, int *NumHotkeys )
{
	if ( !nKey || !Hotkeys || !NumHotkeys )
		return -1;


	int k			= -1;


	for ( int i = 0; i < *NumHotkeys; i++ )
		if ( Hotkeys[i].nKey == nKey )
			k = i;

	if ( k < 0 ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Hotkey not found\r\n" );
		return 1;
	}

	for ( int i = k; i < ( *NumHotkeys - 1 ); i++ )
		Hotkeys[i] = Hotkeys[i + 1];

	*NumHotkeys -= 1;
	return 0;
}



int AddHotkey( int nKey, unsigned short int Action, _Hotkey *Hotkeys, int *NumHotkeys )
{
	if ( !nKey || !Hotkeys || !NumHotkeys )
		return -1;


	if ( !Action )
		return DeleteHotkey( nKey, Hotkeys, NumHotkeys );


	for ( int i = 0; i < *NumHotkeys; i++ ) {
		if ( Hotkeys[i].nKey == nKey ) {
			Hotkeys[i].Action = Action;
			return 0;
		}
	}

	if ( *NumHotkeys >= MAX_KEYS ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Too many hotkeys (%d max)\r\n", MAX_KEYS );
		return 1;
	}

	Hotkeys[*NumHotkeys].nKey		= nKey;
	Hotkeys[*NumHotkeys].Action		= Action;
	*NumHotkeys += 1;
	return 0;
}



DLLExports int WINAPI editkeys( LPTSTR inString )
{
	if ( !inString )
		return 1;


	wchar_t shortbuf[SHORT_BUF_LEN]			= L"";

	HKEY SaveKey				= NULL;
	HKEY RestoreKey				= NULL;
	BOOL Restore				= FALSE;
	BOOL PluginDefaults			= FALSE;
	BOOL ListKeys				= FALSE;
	BOOL Quiet					= FALSE;

	_Hotkey NewHotkeys[MAX_KEYS];
	int NewNumHotkeys			= NumHotkeys;
	BOOL HotkeysChanged			= FALSE;
	int NewBeepMode				= -1;
	int NewCopyBeep				= -1;
	int NewLastArgsFlags		= -1;
	int NewScrollyFlags			= -1;
	int NewAutoCloseFlags		= -1;
	BOOL AnyChanges				= FALSE;
	BOOL ScrollyRedraw			= FALSE;

	unsigned int action			= 0;
	unsigned int index			= 0;
	int keyval					= 0;
	wchar_t *ch					= NULL;

	int trv						= 0;
	int rv						= 0;


	memset( NewHotkeys, 0x00, MAX_KEYS * sizeof( _Hotkey ) );

	if ( NumHotkeys )
		memcpy( NewHotkeys, Hotkeys, NumHotkeys * sizeof( _Hotkey ) );


	ParseArgs( inString, PARSE_BREAK_SPACES | PARSE_SLASHES_KLUDGE | PARSE_FORCE_UPPERCASE );


	for ( int i = 0; i < numargs; i++ ) {
		if ( !wcsncmp( arg[i], L"/?", 2 ) )
			return ShowCmdHelp( EditKeysHelpText, ( arg[i][2] == '?' ) );
		else if ( !_wcsnicmp( arg[i], L"/A", 2 ) ) {
			//	_wcsupr_s( arg[i], MAX_ARG_LEN );
			ch = arg[i] + 2;
			if ( IsColonOrEquals( *ch ) )
				ch++;

			if ( *ch == '\0' ) {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Missing LastArgs flags value: \"%s\"\r\n", arg[i] );
				return 1;
			}

			if ( wcschr( ch, '?' ) ) {
				QPuts( (LPTSTR) EditKeys_OptAHelpText );
				return 0;
			}

			if ( ParseInt( ch, (unsigned int*) &NewLastArgsFlags, &ch ) ) {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Bad LastArgs flags value: \"%s\"\r\n", arg[i] );
				return 1;
			}
			else if ( NewLastArgsFlags >= 0x80000000 ) {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Value too large: %u\r\n", NewLastArgsFlags );
				return 1;
			}

			AnyChanges = TRUE;
		}
		else if ( !_wcsnicmp( arg[i], L"/B", 2 ) ) {
			//	_wcsupr_s( arg[i], MAX_ARG_LEN );
			ch = arg[i] + 2;
			if ( IsColonOrEquals( *ch ) )
				ch++;

			if ( *ch == '\0' ) {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Missing auto-close flags value: \"%s\"\r\n", arg[i] );
				return 1;
			}

			if ( wcschr( ch, '?' ) ) {
				QPuts( (LPTSTR) EditKeys_OptBHelpText );
				return 0;
			}

			if ( ParseInt( ch, (unsigned int*) &NewAutoCloseFlags, &ch ) ) {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Bad auto-close flags value: \"%s\"\r\n", arg[i] );
				return 1;
			}
			else if ( NewAutoCloseFlags >= 0x80000000 ) {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Value too large: %u\r\n", NewLastArgsFlags );
				return 1;
			}

			AnyChanges = TRUE;
		}
		else if ( !_wcsnicmp( arg[i], L"/C", 2 ) ) {
			//	_wcsupr_s( arg[i], MAX_ARG_LEN );
			ch = arg[i] + 2;
			if ( IsColonOrEquals( *ch ) )
				ch++;

			if ( *ch == '\0' ) {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Missing copy-chime value: \"%s\"\r\n", arg[i] );
				return 1;
			}

			if ( wcschr( ch, '?' ) ) {
				QPuts( (LPTSTR) EditKeys_OptCHelpText );
				return 0;
			}

			if ( ParseInt( ch, (unsigned int*) &NewCopyBeep, &ch ) ) {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Bad copy-chime value: \"%s\"\r\n", arg[i] );
				return 1;
			}
			else if ( NewCopyBeep > COPY_BEEP_MAX ) {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Copy-chime value must be 0 - %u: %u\r\n", COPY_BEEP_MAX, NewCopyBeep );
				return 1;
			}

			AnyChanges = TRUE;
		}
		else if ( !_wcsicmp( arg[i], L"/D" ) )
			PluginDefaults = TRUE;
		else if ( !_wcsnicmp( arg[i], L"/E", 2 ) ) {
			//	_wcsupr_s( arg[i], MAX_ARG_LEN );
			ch = arg[i] + 2;
			if ( IsColonOrEquals( *ch ) )
				ch++;

			if ( *ch == '\0' ) {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Missing error-sound value: \"%s\"\r\n", arg[i] );
				return 1;
			}

			if ( wcschr( ch, '?' ) ) {
				QPuts( (LPTSTR) EditKeys_OptEHelpText );
				return 0;
			}

			if ( ParseInt( ch, (unsigned int*) &NewBeepMode, &ch ) ) {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Bad error-sound value: \"%s\"\r\n", arg[i] );
				return 1;
			}
			else if ( NewBeepMode > ERROR_BEEP_MAX ) {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Error-sound value must be 0 - %u: %u\r\n", ERROR_BEEP_MAX, NewBeepMode );
				return 1;
			}

			AnyChanges = TRUE;
		}
		else if ( !_wcsnicmp( arg[i], L"/K", 2 ) ) {
			//	_wcsupr_s( arg[i], MAX_ARG_LEN );
			ch = arg[i] + 2;
			if ( IsColonOrEquals( *ch ) )
				ch++;

			if ( *ch == '\0' ) {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Missing action value: \"%s\"\r\n", arg[i] );
				return 1;
			}

			if ( ParseInt( ch, &action, &ch ) ) {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Bad action value: \"%s\"\r\n", arg[i] );
				return 1;
			}
			else if ( action > ACTION_MAX ) {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Action must be 1 - %u: %u\r\n", ACTION_MAX, action );
				return 1;
			}

			if ( !IsColonOrEquals( *ch ) ) {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Expected a colon or equals sign: \"%s\"\r\n", arg[i] );
				return 1;
			}

			ch++;
			if ( *ch == '\0' ) {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Missing key name: \"%s\"\r\n", arg[i] );
				return 1;
			}

			keyval = ParseKeyName( ch );

			if ( keyval == 0 ) {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Unknown key name: \"%s\"\r\n", ch );
				return 1;
			}
			else if ( keyval == -1 ) {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Invalid hotkey: \"%s\"\r\n", ch );
				return 1;
			}
			else if ( keyval == -2 ) {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Key not allowed: \"%s\"\r\n", ch );
				return 1;
			}

			if ( AddHotkey( keyval, action, NewHotkeys, &NewNumHotkeys ) )
				return 1;

			HotkeysChanged = AnyChanges = TRUE;
		}
		else if ( !_wcsicmp( arg[i], L"/L" ) )
			ListKeys = TRUE;
		else if ( !_wcsicmp( arg[i], L"/Q" ) )
			Quiet = TRUE;
		else if ( !_wcsicmp( arg[i], L"/R" ) )
			Restore = TRUE;
		else if ( !_wcsicmp( arg[i], L"/S" ) )
			SaveKey = HKEY_CURRENT_USER;
		else if ( !_wcsicmp( arg[i], L"/S!" ) )
			SaveKey = HKEY_LOCAL_MACHINE;
		else if ( !_wcsnicmp( arg[i], L"/X", 2 ) ) {
			//	_wcsupr_s( arg[i], MAX_ARG_LEN );
			ch = arg[i] + 2;
			if ( IsColonOrEquals( *ch ) )
				ch++;

			if ( *ch == '\0' ) {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Missing key name: \"%s\"\r\n", arg[i] );
				return 1;
			}

			keyval = ParseKeyName( ch );

			if ( keyval == 0 ) {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Unknown key name: \"%s\"\r\n", ch );
				return 1;
			}
			else if ( keyval == -1 ) {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Invalid hotkey: \"%s\"\r\n", ch );
				return 1;
			}
			else if ( keyval == -2 ) {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Key not allowed: \"%s\"\r\n", ch );
				return 1;
			}

			if ( DeleteHotkey( keyval, NewHotkeys, &NewNumHotkeys ) )
				return 1;

			HotkeysChanged = AnyChanges = TRUE;
		}
		else if ( !_wcsnicmp( arg[i], L"/Y", 2 ) ) {
			//	_wcsupr_s( arg[i], MAX_ARG_LEN );
			ch = arg[i] + 2;
			if ( IsColonOrEquals( *ch ) )
				ch++;

			if ( *ch == '\0' ) {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Missing Scrolly flags value: \"%s\"\r\n", arg[i] );
				return 1;
			}

			if ( wcschr( ch, '?' ) ) {
				QPuts( (LPTSTR) EditKeys_OptYHelpText );
				return 0;
			}

			if ( ParseInt( ch, (unsigned int*) &NewScrollyFlags, &ch ) ) {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Bad Scrolly flags value: \"%s\"\r\n", arg[i] );
				return 1;
			}
			else if ( NewScrollyFlags >= 0x80000000 ) {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Value too large: %u\r\n", NewScrollyFlags );
				return 1;
			}

			AnyChanges = TRUE;
		}
		else if ( arg[i][0] == '/' ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Unknown option: \"%s\"\r\n", arg[i] );
			return 1;
		}
		else {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Syntax error: \"%s\"\r\n", arg[i] );
			return 1;
		}
	}


	//		List suported keys if desired :

	if ( ListKeys ) {
		if ( SaveKey || Restore || PluginDefaults || AnyChanges ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Can\u2019t combine /L with other options\r\n" );
			return 1;
		}

		return ListLegalKeys();
	}


	//		Restore settings if desired :

	if ( Restore ) {
		RestoreKey = RestoreOptions();
		if ( RestoreKey )
			PluginDefaults = FALSE;

		if ( !Quiet ) {
			if ( RestoreKey )
				Printf( L"Settings restored from the registry (%s\\%s).\r\n", RootKeyName( RestoreKey ), PluginRegKey );
			else
				Printf( L"Settings not found in the registry.\r\n" );
		}
	}

	if ( PluginDefaults ) {
		RestorePluginDefaults();

		if ( !Quiet )
			Printf( L"Default plugin settings restored.\r\n" );
	}


	//		Make any requested changes :

	if ( HotkeysChanged ) {
		NumHotkeys = 0;

		memset( Hotkeys, 0x00, MAX_KEYS * sizeof( _Hotkey ) );
		
		if ( NewNumHotkeys )
			memcpy( Hotkeys, NewHotkeys, NewNumHotkeys * sizeof( _Hotkey ) );

		NumHotkeys = NewNumHotkeys;
	}

	if ( NewBeepMode >= 0 )
		BeepMode = NewBeepMode;

	if ( NewCopyBeep >= 0 )
		CopyBeep = NewCopyBeep;

	if ( NewLastArgsFlags >= 0 )
		LastArgsFlags = NewLastArgsFlags;

	if ( NewScrollyFlags >= 0 ) {
		ScrollyRedraw = ( ScrollyFlags & SCROLLY_FLAG_VISIBLE_MARKS ) != ( NewScrollyFlags & SCROLLY_FLAG_VISIBLE_MARKS );
		ScrollyFlags = NewScrollyFlags;

		if ( ScrollyRedraw )
			RedrawScrollyMarks();
	}

	if ( NewAutoCloseFlags >= 0 )
		AutoCloseFlags = NewAutoCloseFlags;


	//		Display the current settings :

	if ( !Quiet ) {
		Printf( L"\r\n%s v%u.%u.%u settings:\r\n\n", PLUGIN_NAME, VER_MAJOR, VER_MINOR, VER_BUILD );

		ShowHotkeyAssignments();

		Printf( L"  Error sound  . . . . . . . . . /E   : %-2d  (%s)\r\n", BeepMode, ErrorBeepName[BeepMode] );
		Printf( L"  Copy chime . . . . . . . . . . /C   : %-2d  (%s)\r\n", CopyBeep, CopyBeepName[CopyBeep] );
		Printf( L"  LastArgs flags . . . . . . . . /A   : %-2u  (%s)\r\n", LastArgsFlags, LastArgsStr( LastArgsFlags ) );
		Printf( L"  Auto-close flags . . . . . . . /B   : %-2d  (%s)\r\n", AutoCloseFlags, AutoCloseStr( AutoCloseFlags ) );
		Printf( L"  Scrolly flags  . . . . . . . . /Y   : %-2u  (%s)\r\n", ScrollyFlags, ScrollyFlagsStr( ScrollyFlags ) );

		Printf( L"\r\n  %d %s defined.\r\n", NumHotkeys, ( NumHotkeys == 1 ) ? L"key" : L"keys" );
	}


	//		Save the current settings if desired :

	if ( SaveKey ) {
		if ( trv = SaveOptions( SaveKey ) ) {
			GetErrorText( trv, shortbuf );
			CrLf();
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Error saving settings to the registry (%s\\%s) :\r\n", RootKeyName( SaveKey ), PluginRegKey );
			Qprintf( ERRHANDLE, L"%s\r\n", shortbuf );
			rv = 2;
		}
		else if ( !Quiet )
			Printf( L"\r\nSettings saved to the registry (%s\\%s).\r\n", RootKeyName( SaveKey ), PluginRegKey );
	}


	return rv;
}



DLLExports int WINAPI key( LPKEYINFO24 ki )
{
	int inKey					= ki->nKey;
	int ShiftKeys				= 0;
	unsigned short int Action	= 0;


	if ( !inKey && ( ki->pszLine == NULL ) )
		UnicodeEntry = FALSE;

	if ( NewKeyHook ) {
		if ( !inKey && ( ki->pszKey ) ) {
			inKey = ParseKeyName24( ki->pszKey );

			if ( inKey <= 0 ) {
				if ( !_wcsicmp( ki->pszKey, L"Enter" ) )
					inKey = 0x0d;
				else if ( !_wcsicmp( ki->pszKey, L"Esc" ) )
					inKey = 0x1b;
			}

			if ( inKey <= 0 )
				return 0;
		}
	}


	if ( GetAsyncKeyState( VK_SHIFT ) & 0x8000 )
		ShiftKeys |= 1;

	if ( GetAsyncKeyState( VK_CONTROL ) & 0x8000 )
		ShiftKeys |= 2;

	if ( GetAsyncKeyState( VK_MENU ) & 0x8000 )
		ShiftKeys |= 4;

	if ( ShiftKeys == 3 )
		inKey |= 0x20000;
	else if ( ShiftKeys == 6 )
		inKey |= 0x40000;
	else if ( NewKeyHook && ( ShiftKeys == 7 ) )
		inKey |= 0x60000;


	if ( UnicodeEntry && ( ( ki->nKey & FBIT ) == 0 ) )
		if ( iswprint( ki->nKey ) && !iswxdigit( ki->nKey ) )
			EndUnicodeEntry( ki, FALSE );


	if ( inKey == 0x0d ) {
		if ( UnicodeEntry )
			EndUnicodeEntry( ki, FALSE );

		MarkStartOfLine( ki->nHomeRow );
		//	SaveLastCommand( ki );

		if ( ki->pszLine[0] == ':' )
			if ( DoConsoleMath( ki ) > 0 )
				return 0;

		return SaveCmdArgs( ki );
	}
	else if ( UnicodeEntry && ( ( inKey == 0x1b ) || ( inKey == 0x1004f ) ) ) {
		EndUnicodeEntry( ki, TRUE );
		return 0;
	}


	if ( inKey == 0x1b ) {

#ifdef CHARINFO_VER

		if ( CharInfoPopup )
			CharInfo_ClosePopup();

#endif	//	CHARINFO_VER

	}


	if ( AutoCloseFlags )
		if ( ( inKey == '[' ) || ( inKey == '(' ) || ( inKey == '{' ) || ( inKey == '\"' ) || ( inKey == '\'' ) || ( inKey == '`' ) )
			AutoClose( ki );


	for ( int i = 0; i < NumHotkeys; i++ ) {
		if ( inKey == Hotkeys[i].nKey ) {
			Action = Hotkeys[i].Action;
			break;
		}
	}

	if ( !Action )
		return 0;


	switch ( Action ) {

	case ACTION_SWAP_CHARS :
		return SwapTwoCharacters( ki );

	case ACTION_SWAP_ARGS :
		return SwapArgs( ki );

	case ACTION_INSERT_LAST_ARGS :
		return InsertLastArgs( ki );

	case ACTION_CLEAR_SCREEN :
		return ClearScreen( ki );

	case ACTION_MATH_WORD :
		return MathWord( ki );

	case ACTION_EXPAND_EXPR :
		return ExpandExpression( ki );

	case ACTION_UNICODE_ENTRY :
		return BeginUnicodeEntry( ki );

	case ACTION_COPY_LINE :
		return CopyCommandLine( ki );

	case ACTION_COPY_WORD :
		return CopyCommandWord( ki, FALSE );

	case ACTION_CUT_WORD :
		return CopyCommandWord( ki, TRUE );

	case ACTION_SWAP_CLIP_WORD :
		return SwapClipWord( ki );

	case ACTION_CHANGE_WORD_CASE :
		return ChangeWordCase( ki );

	case ACTION_CUSTOM_INSERT_1 :
		return CustomInsert( ki, 0 );

	case ACTION_CUSTOM_INSERT_2 :
		return CustomInsert( ki, 1 );

	case ACTION_CUSTOM_INSERT_3 :
		return CustomInsert( ki, 2 );

	case ACTION_CHAR_INFO :
		return ShowCharInfo( ki );

	case ACTION_SCROLLY_UP :
		return ScrollyUp( ki );

	case ACTION_SCROLLY_DOWN :
		return ScrollyDown( ki );

	case ACTION_FILE_URI :
		return FileUriKey( ki );

	case ACTION_PRINT_VIEWPORT :
		return PrintViewport( ki, FALSE );

	case ACTION_SAVE_PICTURE :
		return PrintViewport( ki, TRUE );

	case ACTION_PRINT_HISTORY :
		return PrintHistory( ki );

	case ACTION_PATH_EXPAND :
		return PathExpand( ki );

	default :
		return 0;
	}
}



BOOL APIENTRY DllMain( HANDLE hModule, DWORD  dwReason, LPVOID lpReserved )
{
	switch( dwReason )	{
		case DLL_PROCESS_ATTACH:
			MyDLLHandle = (HINSTANCE) hModule;
			break;

		case DLL_THREAD_ATTACH:
		case DLL_THREAD_DETACH:
		case DLL_PROCESS_DETACH:
			break;
	}
	return TRUE;
}



DLLExports LPPLUGININFO WINAPI GetPluginInfo( void )
{
	static PLUGININFO piInfo;

	piInfo.pszDll			= (LPTSTR) PLUGIN_NAME;
	piInfo.pszAuthor		= (LPTSTR) L"Charles Dye";
	piInfo.pszEmail			= (LPTSTR) L"cdye@unm.edu";
	piInfo.pszWWW			= (LPTSTR) L"http://charlesdye.net/plugins/editkeys.html";
	piInfo.pszDescription	= (LPTSTR) L"EditKeys plugin";
	piInfo.pszFunctions		= (LPTSTR) PluginFeatures;
	piInfo.nMajor			= VER_MAJOR;
	piInfo.nMinor			= VER_MINOR;
	piInfo.nBuild			= VER_BUILD;

	return &piInfo;
}



DLLExports BOOL WINAPI InitializePlugin( void )
{
	wchar_t shortbuf[SHORT_BUF_LEN] = L"%[_transient]%[_pipe]%[noloadmsg]";
	int rv		= 0;

	ConsoleHandle = CreateFile( L"CONOUT$", GENERIC_READ | GENERIC_WRITE, FILE_SHARE_READ | FILE_SHARE_WRITE, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL );

	if ( ( ConsoleHandle == INVALID_HANDLE_VALUE ) || !ConsoleHandle ) {
		if ( !PipeShell() ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Error opening console buffer!\r\n" );
			Qprintf( ERRHANDLE, L"%s plugin not loaded.\r\n", PLUGIN_NAME );
		}

		return -1;
	}

#ifdef NEED_GDIPLUS

	rv = GdiplusStartup( &GdiplusToken, &StartupInput, NULL );
	if ( rv ) {
		if ( !PipeShell() ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Error %d initializing GDI+ \u2014 %s plugin not loaded.\r\n", rv, PLUGIN_NAME );
			CloseHandle( ConsoleHandle );
		}

		return -1;
	}

#endif	//	NEED_GDIPLUS

	GetVersion2( &OsVerInfo );
	InitHotkeys();
	ScrollyInit();
	//	CreatePluginTempDir();
	CharInfo_Initialize();

	RestorePluginDefaults();
	RestoreOptions();


	ExpandVariables( shortbuf, 1 );

	if ( !wcscmp( shortbuf, L"00" ) )
		Printf( L"%s v%u.%u.%u loaded.  %d %s defined.\r\n", PLUGIN_NAME, VER_MAJOR, VER_MINOR, VER_BUILD, NumHotkeys, ( NumHotkeys == 1 ) ? L"key" : L"keys" );

	return 0;
}



DLLExports BOOL WINAPI ShutdownPlugin( BOOL bEndProcess )
{
	free( LastArgs );
	LastArgs = NULL;

	/*
	free( LastCmd );
	LastCmd = NULL;
	*/

	CharInfo_Shutdown();

#ifdef NEED_GDIPLUS
	GdiplusShutdown( GdiplusToken );
#endif	//	NEED_GDIPLUS

	CloseHandle( ConsoleHandle );
	RemovePluginTempDir();

#ifdef NEWHELP_VERSION
	CloseHelpWindow();
#endif

	return 0;
}



const int NumHelpTexts							= 1;
const wchar_t HelpFeature[NumHelpTexts][10]		= { L"editkeys" };
const wchar_t *HelpText[NumHelpTexts]			= { EditKeysHelpText };



#ifndef	NEWHELP_VERSION

LPTSTR UsageText( const wchar_t *inString )
{
	if ( !inString )
		return NULL;


	wchar_t *ch = wcschr( (wchar_t*) inString, '\n' );


	if ( !ch )
		return NULL;

	while ( *ch == '\n' )
		ch++;

	return ch;
}

#endif	//	NEWHELP_VERSION




DLLExports int WINAPI Usage( LPTSTR inString, LPTSTR outString )
{
	if ( ( inString[0] == '@' ) || ( inString[0] == '_' ) )
		return 0;

	for ( int i = 0; i < NumHelpTexts; i++ ) {
		if ( !_wcsicmp( inString, HelpFeature[i] ) ) {
			wcscpy_s( outString, SHORT_BUF_LEN, UsageText( HelpText[i] ) );
			return 1;
		}
	}

	return 0;
}


#ifndef NEWHELP_VERSION

//		Simpler MessageBox() help, useful for smaller plugins which don't merit a full HTML help system:

DLLExports int WINAPI Help( LPTSTR inString )
{
	wchar_t WindowTitle[SHORT_BUF_LEN]		= L"";


	wsprintf( WindowTitle, L"%s Plugin Help", PLUGIN_NAME );

	for ( int i = 0; i < NumHelpTexts; i++ ) {
		if ( !_wcsicmp( inString, HelpFeature[i] ) ) {
			MessageBox( GetForegroundWindow(), HelpText[i], WindowTitle, MB_OK );
			return 1;
		}
	}

	return 0;
}

#endif		//	NEWHELP_VERSION
