
//		ExpandExpression.cpp  --	Functions to find and evaluate a variable,
//									function or expression at the command line.


#ifndef EXPAND_EXPRESSION_VER
#define EXPAND_EXPRESSION_VER			1.0.5.0


#define EE_PARSING_MAIN					0		//	parsing the main command line
#define EE_PARSING_FUNCTION				1		//	%@FUNC[arg,arg,...]
#define EE_PARSING_ARRAY				2		//	%ARRAY[i,j,...]
#define EE_PARSING_EVAL					3		//	%((expr))
#define EE_PARSING_EXEC_RC				4		//	%(command)
#define EE_PARSING_CONDITIONAL			5		//	%[[condition]]
#define EE_PARSING_INDIRECT_VAR			6		//	%[varname]
#define EE_PARSING_EXEC_STRING			7		//	%{command}



int EE_ExpandVariablesQuietly( LPTSTR String )
{
	HANDLE OldStderrHandle	= GetStdHandle( STD_ERROR_HANDLE );
	HANDLE OldStdoutHandle	= GetStdHandle( STD_OUTPUT_HANDLE );
	HANDLE NewStderrHandle	= CreateFile( L"nul", GENERIC_READ | GENERIC_WRITE, FILE_SHARE_READ, NULL, OPEN_ALWAYS, 0, NULL );
	int rv		= 0;


	if ( NewStderrHandle != INVALID_HANDLE_VALUE ) {
		SetStdHandle( STD_ERROR_HANDLE, NewStderrHandle );		//  redirect stderr to nul:
		SetStdHandle( STD_OUTPUT_HANDLE, NewStderrHandle );		//  redirect stdout to nul:
	}
	
	rv = ExpandVariables( String, 1 );							//  and expand the expression

	if ( NewStderrHandle != INVALID_HANDLE_VALUE ) {
		SetStdHandle( STD_ERROR_HANDLE, OldStderrHandle );		//  put stderr back where we found it
		SetStdHandle( STD_OUTPUT_HANDLE , OldStdoutHandle );	//  put stdout back where we found it
		CloseHandle( NewStderrHandle );
	}

	return rv;
}



wchar_t EE_GetEscapeChar()
{
	wchar_t temp[SHORT_BUF_LEN]		= L"%@option[EscapeChar]";


	ExpandVariables( temp, 1 );
	
	if ( temp[0] && ( temp[1] == '\0' ) )
		if ( !iswalnum( temp[0] ) )
			return temp[0];

	return '^';
}



int EE_FindExpression( LPCTSTR String, int Start, int StartArgs, int Cursor, int Depth, int ParsingType, wchar_t Escape, int *outEnd, int *outBestStart, int *outBestEnd, int *outBestDepth )
{
	BOOL quote				= FALSE;
	BOOL escape				= FALSE;
	BOOL escapepvar			= FALSE;
	int pats				= 0;
	int start				= -1;
	int end					= -1;
	int vt					= 0;
	int sa					= -1;
	const wchar_t *v		= NULL;



	for ( const wchar_t *ch = String + StartArgs; *ch; ch++ ) {

		//		Handle escape characters and double quotes :

		if ( *ch == Escape )
			escape = !escape;
		else if ( !escape && ( *ch == QUOTE ) )
			quote = !quote;

		escapepvar = FALSE;

		
		//		Check for end of expression :

		if ( Depth && !quote && !escape ) {

			switch ( ParsingType ) {

			case EE_PARSING_CONDITIONAL :

				if ( ( ch[0] == ']' ) && ( ch[1] == ']' ) )
					end = (int) ( ch - String + 1 );

				break;

			case EE_PARSING_EVAL :

				if ( !pats && ( ch[0] == ')' ) && ( ch[1] == ')' ) )
					end = (int) ( ch - String + 1 );

				break;

			case EE_PARSING_EXEC_RC :

				if ( !pats && ( ch[0] == ')' ) )
					end = (int) ( ch - String );

				break;

			case EE_PARSING_FUNCTION :
			case EE_PARSING_ARRAY :
			case EE_PARSING_INDIRECT_VAR :

				if ( *ch == ']' )
					end = (int) ( ch - String );

				break;

			case EE_PARSING_EXEC_STRING :

				if ( *ch == '}' )
					end = (int) ( ch - String );

				break;
			}
		}

		if ( end >= 0 ) {
			if ( outEnd )
				*outEnd = end;

			if ( ( Start <= Cursor ) && ( end >= ( Cursor - 1 ) ) && ( Depth > *outBestDepth ) ) {
				*outBestStart	= Start;
				*outBestEnd		= end;
				*outBestDepth	= Depth;
			}

			return 1;
		}


		//		Check for start of expression :

		if ( !escape && ( *ch == '%' ) ) {

			vt = 0;
			sa = 0;

			if ( ch[1] == '%' )
				ch++;
			else if ( ( ch[1] == '@' ) && iswalpha( ch[2] ) ) {
				v = ch + 2;

				while ( __iswcsym( *v ) || ( *v == '$' ) )
					v++;

				if ( ( v[0] == '[' ) && v[1] ) {
					vt = EE_PARSING_FUNCTION;
					sa = (int) ( v - String + 1 );
				}
			}
			else if ( __iswcsym( ch[1] ) ) {
				v = ch + 1;
			
				while ( __iswcsym( *v ) || ( *v == '$' ) )
					v++;

				if ( ( v[0] == '[' ) && v[1] ) {
					vt = EE_PARSING_ARRAY;
					sa = (int) ( v - String + 1 );
				}
				else {		//		Found a variable name :
					start	= (int) ( ch - String );
					end		= (int) ( v - String - 1 );
					
					if ( ( start <= Cursor ) && ( end >= ( Cursor - 1 ) ) && ( ( Depth + 1 ) > *outBestDepth ) ) {
						*outBestStart	= start;
						*outBestEnd		= end;
						*outBestDepth	= Depth + 1;
					}

					start = end = -1;
				}
			}
			else if ( ch[1] == '(' ) {
				if ( ( ch[2] == '(' ) && ch[3] ) {
					vt = EE_PARSING_EVAL;
					sa = (int) ( ch - String + 3 );
				}
				else if ( ch[2] ) {
					vt = EE_PARSING_EXEC_RC;
					sa = (int) ( ch - String + 2 );
				}
			}
			else if ( ch[1] == '[' ) {
				if ( ( ch[2] == '[' ) && ch[3] ) {
					vt = EE_PARSING_CONDITIONAL;
					sa = (int) ( ch - String + 3 );
				}
				else if ( ch[2] ) {
					vt = EE_PARSING_INDIRECT_VAR;
					sa = (int) ( ch - String + 2 );
				}
			}
			else if ( ( ch[1] == '{' ) && ch[2] ) {
				vt = EE_PARSING_EXEC_STRING;
				sa = (int) ( ch - String + 2 );
			}
			else if ( ch[1] == '=' ) {
				if ( ( ( ch - String ) <= Cursor ) && ( ( ch - String + 1 ) >= ( Cursor - 1 ) ) && ( ( Depth + 1 ) > *outBestDepth ) ) {
					*outBestStart	= (int) ( ch - String );
					*outBestEnd		= (int) ( ch - String + 1 );
					*outBestDepth	= Depth + 1;
				}

				escape = escapepvar = TRUE;
				ch++;
			}
			else if ( ch[1] == '+' ) {
				if ( ( ( ch - String ) <= Cursor ) && ( ( ch - String + 1 ) >= ( Cursor - 1 ) ) && ( ( Depth + 1 ) > *outBestDepth ) ) {
					*outBestStart	= (int) ( ch - String );
					*outBestEnd		= (int) ( ch - String + 1 );
					*outBestDepth	= Depth + 1;
				}
			}


			if ( vt ) {
				end = -1;
				EE_FindExpression( String, (int) ( ch - String ), sa, Cursor, Depth + 1, vt, Escape, &end, outBestStart, outBestEnd, outBestDepth );

				if ( end >= 0 )
					ch = String + end;

				end = -1;
			}
		}


		//		Handle other characters :

		if ( *ch == '(' )
			pats++;
		else if ( pats && ( *ch == ')' ) )
			pats--;

		if ( ( *ch != Escape ) && !escapepvar )
			escape = FALSE;
	}

	return 0;
}



int ExpandExpression( KEYINFO24 *ki )
{
	if ( !ki )
		goto Error;
	else if ( ki->pszLine == NULL )
		goto Error;
	else if ( ki->pszLine[0] == '\0' )
		goto Error;


	wchar_t VarBuf[MAX_ARG_LEN]			= L"";

	int Start			= -1;
	int End				= -1;
	int Depth			= -1;


	EE_FindExpression( ki->pszLine, 0, 0, (int) ( ki->pszCurrent - ki->pszLine ), 0, EE_PARSING_MAIN, EE_GetEscapeChar(), NULL, &Start, &End, &Depth );

	if ( ( Start < 0 ) || ( End < 0 ) )
		goto Error;


	wcsncpy_s( VarBuf, MAX_ARG_LEN, ki->pszLine + Start, End - Start + 1 );
	//	FmtMsgBox( 0, PLUGIN_NAME, L"Found expression: '%s'", VarBuf );

	if ( EE_ExpandVariablesQuietly( VarBuf ) )
		goto Error;

	//	FmtMsgBox( 0, PLUGIN_NAME, L"New value: '%s'", VarBuf );

	CmdLineDelete( ki, Start, End - Start + 1 );
	CmdLineInsert( ki, VarBuf, Start );
	

	ki->nKey = 0;
	if ( NewKeyHook )
		ki->pszKey[0] = '\0';

	return 0;


Error:

	ErrorBeep();

	ki->nKey = 0;
	if ( NewKeyHook )
		ki->pszKey[0] = '\0';

	return 0;
}


#endif	//	EXPAND_EXPRESSION_VER


/*

		1.0.0.0		2022-05-12
					Rewrote it from scratch to support funky stuff like %((math)) and
					%[[conditional]].

		1.0.1.0		2022-05-16
					Use ExpandVariablesQuietly()

		1.0.2.0		2022-05-24
					Moved GetEscapeChar() into ExpandExpression.cpp.  Added code to
					expand %= and %+ pseudovariables.

		1.0.3.0		2022-05-24
					Moved ExpandVariablesQuietly() in as well.

		1.0.4.0		2022-06-28
					Added support for %(command)

		1.0.5.0		2022-10-13
					Added support for %{command}  (new syntax in TCC v29)

*/
