
//		FileUriKey.cpp


#ifndef FILE_URI_KEY_VER
#define FILE_URI_KEY_VER			1.0.0.0


#include <shlwapi.h>



int FURI_CopyDoublingPercents( LPTSTR inString, LPTSTR outString, size_t BufferSize )
{
	if ( !inString || !outString )
		return -1;


	size_t i = 0, j = 0;
	wchar_t ch = '\0';
	int k = 0;


	do {
		ch = inString[i++];

		if ( j < ( BufferSize - 1 ) ) {
			outString[j++] = ch;

			if ( ch == '%' ) {
				if ( j < ( BufferSize - 1 ) )
					outString[j++] = ch;
				else
					outString[j++] = '\0';

				k++;
			}
		}
	} while ( ch );

	return k;
}



int FURI_UndoublePercents( LPTSTR String )
{
	if ( !String )
		return -1;
	else if ( String[0] == '\0' )
		return 0;


	int j			= 0;
	int k			= 0;


	for ( wchar_t *ch = String; *ch; ch++ ) {
		if ( *ch == '%' )
			if ( ch[1] == '%' )
				ch++, k++;

		String[j++] = *ch;
	}

	String[j] = '\0';
	return k;
}



int FURI_ProblemCharacters( LPTSTR inString )
{
	if ( !inString )
		return 0;


	int k = 0;


	for ( size_t i = 0; inString[i]; i++ ) {
		if ( iswspace( inString[i] ) || iswcntrl( inString[i] ) || ( inString[i] == '<' ) || ( inString[i] == '>' ) || ( inString[i] == '|' ) || ( inString[i] == ',' ) || ( inString[i] == ';' ) || ( inString[i] == '&' ) || ( inString[i] == '%' ) )
			k++;
	}
	
	return k;
}



int FileToUri( LPTSTR Filename )
{
	if ( !Filename )
		return -1;

	
	wchar_t Buffer[MAX_ARG_LEN]		= L"";
	DWORD BufferSize				= MAX_ARG_LEN;
	HRESULT rv						= S_OK;


	TrimLine( Filename );
	if ( Filename[0] == '\0' )
		return -1;

	QueryTrueName( Filename, Buffer );
	wcscpy_s( Filename, MAX_ARG_LEN, Buffer );

	rv = UrlCreateFromPath( Filename, Buffer, &BufferSize, NULL );

	if ( ( rv != S_OK ) && ( rv != S_FALSE ) )
		return -1;

	FURI_CopyDoublingPercents( Buffer, Filename, MAX_ARG_LEN );
	return 0;
}



int FileUriKey( KEYINFO24 *ki )
{
	wchar_t Filename[MAX_FILENAME_LEN]		= L"";
	wchar_t temp[MAX_FILENAME_LEN]			= L"";

	size_t WordStart		= 0;
	size_t WordLen			= 0;
	DWORD BufferSize		= 0;
	BOOL NeedQuotes			= FALSE;


	ki->nKey = 0;
	if ( NewKeyHook )
		ki->pszKey[0] = '\0';


	if ( !FindWordAtCursor( ki, &WordStart, &WordLen, Filename, MAX_FILENAME_LEN ) )
		goto Fail;


	if ( _wcsnicmp( Filename, L"file://", 7 ) ) {
		if ( FileToUri( Filename ) )
			goto Fail;
	}
	else {
		temp[0] = '\0';
		BufferSize = MAX_ARG_LEN;

		FURI_UndoublePercents( Filename );
		PathCreateFromUrl( Filename, temp, &BufferSize, 0 );

		if ( temp[0] == '\0' )
			goto Fail;

		if ( FURI_ProblemCharacters( temp ) ) {
			Filename[0] = QUOTE;
			Filename[1] = '\0';
			wcscat_s( Filename, MAX_FILENAME_LEN, temp );
			wcscat_s( Filename, MAX_FILENAME_LEN, L"\"" );
		}
		else
			wcscpy_s( Filename, MAX_FILENAME_LEN, temp );
	}


	//	FmtMsgBox( 0, PLUGIN_NAME, L"Filename = %s", Filename );

	CmdLineDelete( ki, WordStart, WordLen );
	CmdLineInsert( ki, Filename, WordStart );
	ki->pszCurrent = ki->pszLine + WordStart + wcslen( Filename );
	return 0;


Fail:

	ErrorBeep();
	return 0;
}


#endif	//	FILE_URI_KEY_VER


/*

	1.0.0.0		2022-07-05

*/
