
//		PathExpand.cpp


#ifndef PATH_EXPAND_VER
#define PATH_EXPAND_VER								1.0.1.1


#define PATH_SEARCH_FLAG_SEARCH_HERE_FIRST			0x0001
#define PATH_SEARCH_FLAG_APP_PATHS					0x0002
#define PATH_SEARCH_FLAG_EXECUTABLE_EXTENSIONS		0x0004
#define PATH_SEARCH_FLAG_FILENAME_ONLY				0x0040
#define PATH_SEARCH_FLAG_REG_EXPAND					0x0080



const wchar_t PathExp_FlagsVarName[]			= L"PathExpandFlags";



int PathExp_IsExecutable( LPCTSTR Filename, unsigned int Flags )
{
	if ( !Filename )
		return 0;


	wchar_t Ext[SHORT_BUF_LEN]			= L"";


	_ExtensionPart( (LPTSTR) Filename, Ext );


	if ( Ext[0] != '.' )
		return 0;

	_wcslwr_s( Ext, SHORT_BUF_LEN );

	if ( !wcscmp( Ext, L".exe" ) )
		return 1;

#ifndef _WIN64

	if ( !wcscmp( Ext, L".com" ) || !wcscmp( Ext, L".pif" ) )
		return 2;

#endif	//	_WIN64

	if ( !wcscmp( Ext, L".bat" ) || !wcscmp( Ext, L".cmd" ) || !wcscmp( Ext, L".btm" ) )
		return 3;


	if ( Flags & PATH_SEARCH_FLAG_EXECUTABLE_EXTENSIONS ) {
		wchar_t Value[MAX_ARG_LEN]			= L"";

		GetEnvironmentVariable( Ext, Value, MAX_ARG_LEN );

		if ( Value[0] )
			return 4;
	}

	return 0;
}



int PathExp_SearchAppPaths( HKEY RootKey, LPCTSTR WildSpec, LPTSTR outFilename, size_t buflen, BOOL Expand )
{
	if ( !RootKey || !WildSpec || !outFilename || ( buflen < MAX_PATH ) )
		return -1;


	wchar_t SubKey[SHORT_BUF_LEN]			= L"";
	wchar_t Value[MAX_FILENAME_LEN]			= L"";

	HKEY KeyHandle			= NULL;
	HKEY SubKeyHandle		= NULL;
	DWORD i					= 0;
	LPCTSTR Ext				= NULL;
	DWORD ValType			= 0;
	DWORD ValSize			= 0;


	if ( RegOpenKeyEx( RootKey, L"SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\App Paths", 0, KEY_READ, &KeyHandle ) )
		return 0;

	do {
		SubKey[0] = '\0';
		if ( RegEnumKey( KeyHandle, i, SubKey, SHORT_BUF_LEN ) == ERROR_NO_MORE_ITEMS )
			break;

		i++;
		if ( SubKey[0] == '\0' )
			continue;

		if ( WildcardComparison( (LPTSTR) WildSpec, SubKey, 1, 1 ) )
			continue;

		if ( RegOpenKeyEx( KeyHandle, SubKey, 0, KEY_READ, &SubKeyHandle ) == ERROR_SUCCESS ) {

			Value[0] = '\0';
			ValType = 0;
			ValSize = MAX_FILENAME_LEN * sizeof( wchar_t );

			if ( RegQueryValueEx( SubKeyHandle, NULL, NULL, &ValType, (BYTE*) Value, &ValSize ) )
				Value[0] = '\0';
			else if ( ( ValType != REG_SZ ) && ( ValType != REG_EXPAND_SZ ) )
				Value[0] = '\0';

			RegCloseKey( SubKeyHandle );
			SubKeyHandle = NULL;
		}

		if ( Value[0] ) {

 			if ( Expand && ( ValType == REG_EXPAND_SZ ) && wcschr( Value, '%' ) ) {
				wchar_t temp[MAX_FILENAME_LEN]		= L"";

				if ( ExpandEnvironmentStrings( Value, temp, MAX_FILENAME_LEN ) )
					if ( temp[0] )
						wcscpy_s( Value, MAX_FILENAME_LEN, temp );
			}

			Ext = _ExtensionPart( Value, NULL );

			if ( Ext ) {
				if ( !_wcsicmp( Ext, L".exe" ) ) {
					if ( wcschr( Value, '%' ) || IsAFile( Value ) ) {
						//	Found a match!

						wcscpy_s( outFilename, buflen, Value );
						RegCloseKey( KeyHandle );
						return 1;
					}
				}
			}
		}
	}
	while ( TRUE );

	RegCloseKey( KeyHandle );
	return 0;
}



int PathExp_SearchOnePath( LPCTSTR Wildspec, LPCTSTR Dir, unsigned int Flags, LPTSTR outFilename, size_t buflen )
{
	if ( !Wildspec || !Dir || !outFilename || ( buflen < MAX_PATH ) )
		return -1;


	wchar_t SearchSpec[MAX_FILENAME_LEN]		= L"";

	LPCTSTR Ext					= _ExtensionPart( Wildspec, NULL );

	HANDLE SearchHandle			= NULL;
	WIN32_FIND_DATA Found;


	wcscpy_s( SearchSpec, MAX_FILENAME_LEN, Dir );
	AppendFilename( SearchSpec, L"*", MAX_FILENAME_LEN );
	if ( Ext )
		wcscat_s( SearchSpec, MAX_FILENAME_LEN, Ext );


	SearchHandle = FindFirstFile( SearchSpec, &Found );
	if ( !SearchHandle || ( SearchHandle == INVALID_HANDLE_VALUE ) )
		return 0;

	do {
		if ( Found.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY )
			continue;

		if ( WildcardComparison( (LPTSTR) Wildspec, Found.cFileName, 1, 1 ) )
			continue;

		if ( !Ext )
			if ( !PathExp_IsExecutable( Found.cFileName, Flags ) )
				continue;

		//		Found a match!

		wcscpy_s( outFilename, buflen, Dir );
		AppendFilename( outFilename, Found.cFileName, buflen );
		FindClose( SearchHandle );
		return 1;
	}
	while ( FindNextFile( SearchHandle, &Found ) );

	FindClose( SearchHandle );
	return 0;
}



int PathExp_SearchPath( LPCTSTR Wildspec, unsigned int Flags, LPTSTR outFilename, size_t buflen )
{
	if ( !Wildspec || !outFilename || ( buflen < MAX_PATH ) )
		return -1;


	wchar_t PathValue[MAX_ARG_LEN]				= L"";
	wchar_t CurrentDir[MAX_FILENAME_LEN]		= L"";

	size_t j						= 0;
	BOOL quote					= FALSE;
	int rv						= 0;



	if ( Flags & PATH_SEARCH_FLAG_SEARCH_HERE_FIRST ) {
		GetCurrentDirectory( MAX_FILENAME_LEN, CurrentDir );
		if ( CurrentDir[0] ) {
			//	FmtMsgBox( 0, PLUGIN_NAME, L"Search current directory:\n'%s'", CurrentDir );

			rv = PathExp_SearchOnePath( Wildspec, CurrentDir, Flags, outFilename, buflen );
			if ( rv > 0 )
				return rv;
		}
	}


	if ( Flags & PATH_SEARCH_FLAG_APP_PATHS ) {
		//	FmtMsgBox( 0, PLUGIN_NAME, L"Search current user's App Paths:" );

		rv = PathExp_SearchAppPaths( HKEY_CURRENT_USER, Wildspec, outFilename, buflen, ( Flags & PATH_SEARCH_FLAG_REG_EXPAND ) > 0 );
		if ( rv > 0 )
			return rv;

		//	FmtMsgBox( 0, PLUGIN_NAME, L"Search common App Paths:" );

		rv = PathExp_SearchAppPaths( HKEY_LOCAL_MACHINE, Wildspec, outFilename, buflen, ( Flags & PATH_SEARCH_FLAG_REG_EXPAND ) > 0 );
		if ( rv > 0 )
			return rv;
	}


	GetEnvironmentVariable( L"Path", PathValue, MAX_ARG_LEN );
	if ( PathValue[0] ) {
		//	FmtMsgBox( 0, PLUGIN_NAME, L"Search the path:\n%s", PathValue );

		j = 0;
		CurrentDir[j] = '\0';
		quote = FALSE;

		for ( const wchar_t *ch = PathValue; *ch; ch++ ) {
			if ( iswspace( *ch ) && !quote && !j )
				continue;

			if ( *ch == '"' ) {
				quote = !quote;
				continue;
			}

			if ( !quote && ( *ch == ';' ) ) {
				if ( j ) {
					if ( QueryIsDirectory( CurrentDir ) ) {
						rv = PathExp_SearchOnePath( Wildspec, CurrentDir, Flags, outFilename, buflen );
						if ( rv > 0 )
							return rv;
					}
				}

				j = 0;
				CurrentDir[j] = '\0';
			}
			else if ( j < ( MAX_FILENAME_LEN - 1 ) ) {
				CurrentDir[j++] = *ch;
				CurrentDir[j] = '\0';
			}
		}
	}

	return rv;
}



int PathExpand( KEYINFO24 *ki )
{
	wchar_t SearchSpec[MAX_FILENAME_LEN]	= L"";
	wchar_t FoundFile[MAX_FILENAME_LEN]		= L"";
	wchar_t temp[MAX_FILENAME_LEN]			= L"";

	size_t WordStart			= 0;
	size_t WordLen				= 0;

	unsigned int Flags			= EnvVarValue( PathExp_FlagsVarName, PATH_SEARCH_FLAG_APP_PATHS | PATH_SEARCH_FLAG_EXECUTABLE_EXTENSIONS , 0, UINT_MAX );


	ki->nKey = 0;

	if ( NewKeyHook )
		ki->pszKey[0] = '\0';


	if ( !FindWordAtCursor( ki, &WordStart, &WordLen, SearchSpec, MAX_FILENAME_LEN ) )
		goto Fail;

	if ( wcschr( SearchSpec, '/' ) || wcschr( SearchSpec, '\\' ) || wcschr( SearchSpec, ':' ) )
		goto Fail;

	if ( LastChar( SearchSpec ) != '*' )
		wcscat_s( SearchSpec, L"*" );


	if ( PathExp_SearchPath( SearchSpec, Flags, FoundFile, MAX_FILENAME_LEN ) <= 0 ) 
		goto Fail;

	CmdLineDelete( ki, WordStart, WordLen );

	if ( Flags & PATH_SEARCH_FLAG_FILENAME_ONLY ) {
		_FilenamePart( FoundFile, temp );
		AddQuotes( temp );
		CmdLineInsert( ki, temp, WordStart );
	}
	else {
		AddQuotes( FoundFile );
		CmdLineInsert( ki, FoundFile, WordStart );
	}

	return 0;


Fail:

	ErrorBeep();
	return 0;
}



#endif	//	PATH_EXPAND_VER



/*

	1.0.0.0		2022-07-27

	1.0.1.0		2022-08-02
				Added a flag to control the executable-extensions check.

	1.0.1.1		2023-06-13
				Removed a call to QueryIsFile() to work around an issue in
				TCC v30.00.20.

*/
