
//		PluginTempDir.cpp


#ifndef PLUGIN_TEMP_DIR_VER
#define PLUGIN_TEMP_DIR_VER			1.0.0.0



wchar_t *PluginTempDir				= NULL;



int PTD_RemoveDirectory( LPCTSTR Directory )
{
	if ( !Directory )
		return -1;
	else if ( Directory[0] == '\0' )
		return -1;


	wchar_t wild[MAX_FILENAME_LEN]		= L"";
	wchar_t FullName[MAX_FILENAME_LEN]	= L"";
	
	HANDLE SearchHandle			= NULL;
	WIN32_FIND_DATA FoundItem;
	unsigned int Failed			= 0;
	int rv						= 0;


	if ( !QueryIsDirectory( (LPTSTR) Directory ) )
		return 0;

	wcscpy_s( wild, MAX_FILENAME_LEN, Directory );
	AppendFilename( wild, L"*", MAX_FILENAME_LEN );

	SearchHandle = FindFirstFile( wild, &FoundItem );

	if ( SearchHandle && ( SearchHandle != INVALID_HANDLE_VALUE ) ) {

		do {
			if ( !wcscmp( FoundItem.cFileName, L"." ) || !wcscmp( FoundItem.cFileName, L".." ) )
				continue;

			wcscpy_s( FullName, MAX_FILENAME_LEN, Directory );
			AppendFilename( FullName, FoundItem.cFileName, MAX_FILENAME_LEN );

			if ( FoundItem.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY ) {
				if ( PTD_RemoveDirectory( FullName ) <= 0 )
					Failed++;
			}
			else {
				if ( FoundItem.dwFileAttributes & FILE_ATTRIBUTE_READONLY )
					SetFileAttributes( FullName, 0 );

				if ( !DeleteFile( FullName ) )
					Failed++;
			}
		}
		while ( FindNextFile( SearchHandle, &FoundItem ) );

		FindClose( SearchHandle );
	}

	if ( !Failed )
		if ( RemoveDirectory( Directory ) )
			rv = 1;

	return rv;
}



int CreatePluginTempDir()
{
	wchar_t TempPath[MAX_PATH + 1]			= L"";
	wchar_t Working[MAX_FILENAME_LEN]		= L"";

	DWORD MyPid						= GetCurrentProcessId();
	unsigned int idx				= 0;


	if ( PluginTempDir ) {
		if ( QueryIsDirectory( PluginTempDir ) )
			return 0;

		free( PluginTempDir );
		PluginTempDir = NULL;
	}


	GetTempPath( MAX_PATH + 1, TempPath );
	if ( TempPath[0] == '\0' )
		return 0;

	if ( !QueryIsDirectory( TempPath ) )
		return 0;


	wsprintf( Working, L"%s%s-%04X", TempPath, PLUGIN_NAME, MyPid );

	if ( QueryIsFileOrDirectory( Working ) ) {
		rand_s( &idx );

		for ( int i = 0; i < 1000; i++ ) {
			wsprintf( Working, L"%s%s-%04X-%08X", TempPath, PLUGIN_NAME, MyPid, idx );

			if ( !QueryIsFileOrDirectory( Working ) )
				break;

			idx++;
		}
	}

	if ( CreateDirectory( Working, NULL ) ) {
		PluginTempDir = (wchar_t*) malloc( ( wcslen( Working ) + 1 ) * sizeof( wchar_t ) );
		if ( PluginTempDir ) {
			wcscpy_s( PluginTempDir, wcslen( Working ) + 1, Working );
			return 1;
		}
	}

	return 0;
}



int RemovePluginTempDir()
{
	int rv			= 0;


	if ( PluginTempDir ) {
		if ( PTD_RemoveDirectory( PluginTempDir ) )
			rv = 1;

		free( PluginTempDir );
		PluginTempDir = NULL;
	}

	return rv;
}



int CreateTempFilename( LPTSTR Filename, size_t buflen, LPCTSTR Ext, LPCTSTR Prefix )
{
	if ( !Filename || ( buflen < MAX_PATH ) )
		return -1;


	wchar_t temp[MAX_FILENAME_LEN]			= L"";
	wchar_t fn[SHORT_BUF_LEN]				= L"";
	const wchar_t DefaultExt[]				= L".tmp";

	SYSTEMTIME Now;
	unsigned int idx			= 0;


	Filename[0] = '\0';

	if ( !PluginTempDir ) {
		CreatePluginTempDir();
		if ( !PluginTempDir )
			return ERROR_INVALID_PARAMETER;
	}

	if ( !QueryIsDirectory( PluginTempDir ) )
		return ERROR_PATH_NOT_FOUND;

	if ( Prefix )
		if ( Prefix[0] == '\0' )
			Prefix = NULL;

	if ( !Ext )
		Ext = DefaultExt;
	else if ( Ext[0] != '.' )
		Ext = DefaultExt;

	GetLocalTime( &Now );


	if ( Prefix )
		wsprintf( fn, L"%s-%04u%02u%02u-%02u%02u%02u%s", Prefix, Now.wYear, Now.wMonth, Now.wDay, Now.wHour, Now.wMinute, Now.wSecond, Ext );
	else
		wsprintf( fn, L"%04u%02u%02u-%02u%02u%02u%s", Now.wYear, Now.wMonth, Now.wDay, Now.wHour, Now.wMinute, Now.wSecond, Ext );

	wcscpy_s( temp, MAX_FILENAME_LEN, PluginTempDir );
	AppendFilename( temp, fn, MAX_FILENAME_LEN );

	if ( QueryIsFileOrDirectory( temp ) ) {
		rand_s( &idx );

		for ( int i = 0; i < 1000; i++ ) {
			if ( Prefix )
				wsprintf( fn, L"%s-%04u%02u%02u-%02u%02u%02u-%08X%s", Prefix, Now.wYear, Now.wMonth, Now.wDay, Now.wHour, Now.wMinute, Now.wSecond, idx, Ext );
			else
				wsprintf( fn, L"%04u%02u%02u-%02u%02u%02u-%08X%s", Now.wYear, Now.wMonth, Now.wDay, Now.wHour, Now.wMinute, Now.wSecond, idx, Ext );

			wcscpy_s( temp, MAX_FILENAME_LEN, PluginTempDir );
			AppendFilename( temp, fn, MAX_FILENAME_LEN );
			idx += 1007;

			if ( !QueryIsFileOrDirectory( temp ) )
				break;
		}
	}

	if ( QueryIsFileOrDirectory( temp ) )
		return ERROR_ALREADY_EXISTS;


	if ( wcslen( temp ) >= buflen )
		return ERROR_BUFFER_OVERFLOW;

	wcscpy_s( Filename, buflen, temp );
	return 0;
}



int CreateTempFilename( LPTSTR Filename, size_t buflen, LPCTSTR Ext )
{
	return CreateTempFilename( Filename, buflen, Ext, NULL );
}


#endif	//	PLUGIN_TEMP_DIR_VER


/*

	1.0.0.0		2022-07-12

*/
