
//		PrintHistory.cpp

//		Dumps the last 100 lines of the command history to a .TXT file,
//		then prints the file using the default handler.


#ifndef PRINT_HISTORY_VER
#define PRINT_HISTORY_VER						1.0.0.1


#define PRINT_HISTORY_FLAG_LINE_NUMBERS			0x0001
#define PRINT_HISTORY_FLAG_UTF16				0x0010
#define PRINT_HISTORY_FLAG_DEBUG				0x0400

#define PRINT_HISTORY_MAX_LINES_DEFAULT			100
#define PRINT_HISTORY_FLAGS_DEFAULT				0


#include "BufferedOutput.cpp"



const wchar_t PH_MaxLinesVarName[]		= L"PrintHistoryMaxLines";
const wchar_t PH_FlagsVarName[]			= L"PrintHistoryFlags";



LPCTSTR _HistoryStart()
{
	LPCTSTR Line			= HistoryStart();


	if ( Line )
		if ( Line[0] == '\0' )
			return NULL;

	return Line;
}



LPCTSTR _NextHistoryEntry( LPCTSTR Line )
{
	if ( !Line )
		return NULL;


	LPCTSTR Next		= NextHistoryEntry( (LPTSTR) Line );


	if ( Next ) {
		if ( Next[0] == '\0' )
			return NULL;
		else if ( Next == Line )
			return NULL;
	}

 	return Next;
}



int PrintHistory( KEYINFO24 *ki )
{
	wchar_t Filename[MAX_FILENAME_LEN]		= L"";
	wchar_t temp[SHORT_BUF_LEN]				= L"";

	LPCTSTR Line					= _HistoryStart();
	unsigned int Count				= 0;
	unsigned int k					= 0;
	unsigned int MaxLines			= EnvVarValue( PH_MaxLinesVarName, PRINT_HISTORY_MAX_LINES_DEFAULT );
	unsigned int Flags				= EnvVarValue( PH_FlagsVarName, PRINT_HISTORY_FLAGS_DEFAULT );
	const int BufferSize			= 16384;
	unsigned int Encoding			= ( Flags & PRINT_HISTORY_FLAG_UTF16 ) ? BUFOUT_FILE_UTF16 : BUFOUT_FILE_UTF8;

	BufOut_Buffer *OutBuffer		= NULL;



	if ( !MaxLines )
		MaxLines = PRINT_HISTORY_MAX_LINES_DEFAULT;


	ki->nKey = 0;
	if ( NewKeyHook )
		ki->pszKey[0] = '\0';


	while ( Line ) {
		Count++;
		Line = _NextHistoryEntry( Line );
	}

	if ( !Count )
		goto Error;


	Line = _HistoryStart();

	if ( Count > MaxLines ) {
		for ( unsigned int i = 0; i < ( Count - MaxLines ); i++ )
			Line = _NextHistoryEntry( Line ), k++;

		if ( !Line )
			goto Error;
	}


	CreateTempFilename( Filename, MAX_FILENAME_LEN, L".txt" );
	if ( Filename[0] == '\0' )
		goto Error;

	OutBuffer = OpenBufferedFile( Filename, BufferSize, Encoding, 0 );
	if ( !OutBuffer )
		goto Error;

	if ( BufferedOutput( OutBuffer, BOM ) )
		goto Error;

	while ( Line ) {
		k++;

		if ( Flags & PRINT_HISTORY_FLAG_LINE_NUMBERS ) {
			wsprintf( temp, L"%5u:  ", k );
			if ( BufferedOutput( OutBuffer, temp ) )
				goto Error;
		}

		if ( BufferedOutput( OutBuffer, Line ) )
			goto Error;

		if ( BufferedOutput( OutBuffer, L"\n" ) )
			goto Error;

		Line = _NextHistoryEntry( Line );
	}

	if ( CloseBufferedFile( OutBuffer ) )
		goto Error;

	if ( Flags & PRINT_HISTORY_FLAG_DEBUG )
		ShellExecute( NULL, NULL, Filename, NULL, NULL, SW_SHOW );
	else
		ShellExecute( NULL, L"Print", Filename, NULL, NULL, SW_HIDE );

	return 0;


Error:

	if ( OutBuffer )
		CloseBufferedFile( OutBuffer );

	ErrorBeep();
	return 0;
}


#endif	//	PRINT_HISTORY_VER


/*

	1.0.0.0		2022-07-13

	1.0.0.1		2022-07-14
				Use ShellExecute() instead of LaunchFile().

*/
