
//		Scroll backward and foreward by commands


#ifndef SCROLLY_VERSION
#define SCROLLY_VERSION					1.1.0.0


#define SCROLLY_FLAG_WRAP_AROUND		0x0001
#define SCROLLY_FLAG_VISIBLE_MARKS		0x0002
#define CONSOLE_HIGH_ATTRIBUTES			0xff00
#define CONSOLE_LOW_ATTRIBUTES			0x00ff
#define CONSOLE_BACKGROUND				0x00f0
#define SCROLLY_HIGHLIGHT_1				0x001f			//	bright white on blue
#define SCROLLY_HIGHLIGHT_2				0x004e			//	bright yellow on red



//		--------------------  Scrolly global variables  --------------------

DWORD ScrollyFlags				= SCROLLY_FLAG_WRAP_AROUND;
WORD MarkValue					= 0x2000;

//		--------------------------------------------------------------------




int ScrollyInit()
{
	if ( OSVerCmp( &OsVerInfo, 10, 0, 19041 ) >= 0 )
		MarkValue = COMMON_LVB_GRID_LVERTICAL;
	else
		MarkValue = 0x2000;

	return 0;
}



LPCTSTR ScrollyFlagsStr( unsigned int Flags )
{
	static wchar_t temp[128]		= L"";


	temp[0] = '\0';

	if ( Flags & SCROLLY_FLAG_WRAP_AROUND )
		wcscpy_s( temp, SHORT_BUF_LEN, L"Wrap" );

	if ( Flags & SCROLLY_FLAG_VISIBLE_MARKS ) {
		if ( temp[0] )
			wcscat_s( temp, SHORT_BUF_LEN, L", " );

		wcscat_s( temp, SHORT_BUF_LEN, L"Visible" );
	}

	if ( temp[0] == '\0' )
		wcscpy_s( temp, SHORT_BUF_LEN, L"none" );

	return temp;
}



int GetCursorLine()
{
	HANDLE ConsoleHandle						= NULL;
	CONSOLE_SCREEN_BUFFER_INFO BufferInfo;
	int rv										= -1;


	ConsoleHandle = CreateFile( L"CONOUT$", GENERIC_READ | GENERIC_WRITE, FILE_SHARE_READ | FILE_SHARE_WRITE, NULL, OPEN_EXISTING, 0, NULL );
	if ( ConsoleHandle != INVALID_HANDLE_VALUE ) {

		memset( &BufferInfo, 0x00, sizeof( CONSOLE_SCREEN_BUFFER_INFO ) );

		if ( GetConsoleScreenBufferInfo( ConsoleHandle, &BufferInfo ) )
			rv = BufferInfo.dwCursorPosition.Y;

		CloseHandle( ConsoleHandle );
	}
	
	return rv;
}



int MarkStartOfLine( int Line )
{
	HANDLE ConsoleHandle	= NULL;
	WORD Attr				= 0x0000;
	COORD Coord;
	DWORD CellsRead			= 0;
	DWORD CellsWritten		= 0;


	ConsoleHandle = CreateFile( L"CONOUT$", GENERIC_READ | GENERIC_WRITE, FILE_SHARE_READ | FILE_SHARE_WRITE, NULL, OPEN_EXISTING, 0, NULL );

	if ( ConsoleHandle == INVALID_HANDLE_VALUE )
		return 0;


	Coord.X		= 0;
	Coord.Y		= Line;

	if ( ReadConsoleOutputAttribute( ConsoleHandle, &Attr, 1, Coord, &CellsRead ) ) {
		if ( ( ScrollyFlags & SCROLLY_FLAG_VISIBLE_MARKS ) && ( ( Attr & MarkValue ) == 0 ) ) {
			if ( ( Attr & CONSOLE_BACKGROUND ) == BACKGROUND_BLUE )
				Attr = ( Attr & CONSOLE_HIGH_ATTRIBUTES ) | MarkValue | SCROLLY_HIGHLIGHT_2;
			else
				Attr = ( Attr & CONSOLE_HIGH_ATTRIBUTES ) | MarkValue | SCROLLY_HIGHLIGHT_1;
		}
		else
			Attr |= MarkValue;

		WriteConsoleOutputAttribute( ConsoleHandle, &Attr, 1, Coord, &CellsWritten );
	}

	CloseHandle( ConsoleHandle );
	return 0;
}



BOOL WouldNotActuallyScroll( CONSOLE_SCREEN_BUFFER_INFO *BufferInfo, int Line )
{
	int TopLine				= BufferInfo->srWindow.Top;
	int NumLines			= BufferInfo->dwSize.Y;
	int ViewportHeight		= BufferInfo->srWindow.Bottom - BufferInfo->srWindow.Top;


	if ( ( Line < 0 ) || ( Line >= NumLines ) )
		return FALSE;
	else if ( Line >= ( NumLines - ViewportHeight ) )
		Line = NumLines - ViewportHeight - 1;

	if ( Line == TopLine )
		return TRUE;

	return FALSE;
}



int FindNextMark( HANDLE ConsoleHandle, CONSOLE_SCREEN_BUFFER_INFO *BufferInfo, BOOL Up )
{
	const int ChunkSize		= 512;
	int NumLines			= BufferInfo->dwSize.Y;
	int TopLine				= BufferInfo->srWindow.Top;
	int rv					= -1;

	int FirstMark			= -1;
	int PrevMark			= -1;
	int NextMark			= -1;
	int LastMark			= -1;
	unsigned int k			= 0;

	CHAR_INFO ReadBuffer[ChunkSize];
	COORD BufferSize;
	COORD ReadCoord;
	SMALL_RECT ReadRegion;



	BufferSize.X	= 1;
	BufferSize.Y	= ChunkSize;
	ReadCoord.X		= 0;
	ReadCoord.Y		= 0;


	for ( int row = 0; row < NumLines; row++ ) {

		if ( ( row % ChunkSize ) == 0 ) {
			ReadRegion.Top		= row;
			ReadRegion.Bottom	= row + ChunkSize - 1;
			ReadRegion.Left		= 0;
			ReadRegion.Right	= 0;

			if ( ReadRegion.Bottom >= NumLines )
				ReadRegion.Bottom = NumLines - 1;

			if ( !ReadConsoleOutput( ConsoleHandle, ReadBuffer, BufferSize, ReadCoord, &ReadRegion ) )
				return -1;
		}

		if ( ReadBuffer[row % ChunkSize].Attributes & MarkValue ) {

			if ( FirstMark < 0 )
				FirstMark = row;

			if ( row < TopLine )
				PrevMark = row;

			if ( ( row > TopLine ) && ( NextMark < 0 ) )
				NextMark = row;

			LastMark = row;
			k++;
		}
	}


	if ( Up ) {
		rv = PrevMark;

		if ( ( rv < 0 ) && ( ScrollyFlags & SCROLLY_FLAG_WRAP_AROUND ) )
			rv = LastMark;
	}
	else {
		rv = NextMark;

		if ( ( rv < 0 ) && ( ScrollyFlags & SCROLLY_FLAG_WRAP_AROUND ) )
			rv = FirstMark;

		if ( rv > 0 ) {
			if ( WouldNotActuallyScroll( BufferInfo, rv ) ) {
				if ( ScrollyFlags & SCROLLY_FLAG_WRAP_AROUND )
					rv = FirstMark;
				else
					rv = -1;
			}
		}
	}

	return rv;
}



BOOL ScrollToLine( HANDLE ConsoleHandle, CONSOLE_SCREEN_BUFFER_INFO *BufferInfo, int Line )
{
	int NumLines			= BufferInfo->dwSize.Y;
	int ViewportWidth		= BufferInfo->srWindow.Right - BufferInfo->srWindow.Left;
	int ViewportHeight		= BufferInfo->srWindow.Bottom - BufferInfo->srWindow.Top;
	SMALL_RECT SmallRect;

	int rv					= FALSE;


	if ( ( Line < 0 ) || ( Line >= NumLines ) )
		return FALSE;
	else if ( Line >= ( NumLines - ViewportHeight ) )
		Line = NumLines - ViewportHeight - 1;

	SmallRect.Top		= Line;
	SmallRect.Left		= 0;
	SmallRect.Bottom	= Line + ViewportHeight;
	SmallRect.Right		= ViewportWidth;

	if ( SetConsoleWindowInfo( ConsoleHandle, TRUE, &SmallRect ) )
		rv = TRUE;

	return rv;
}



int ScrollToMark( int CurrentLine, BOOL Up )
{
	HANDLE ConsoleHandle		= NULL;
	CONSOLE_SCREEN_BUFFER_INFO BufferInfo;
	
	int NumLines				= 0;
	int NewLine					= -1;
	BOOL Success				= FALSE;



	ConsoleHandle = CreateFile( L"CONOUT$", GENERIC_READ | GENERIC_WRITE, FILE_SHARE_READ | FILE_SHARE_WRITE, NULL, OPEN_EXISTING, 0, NULL );

	if ( ConsoleHandle != INVALID_HANDLE_VALUE ) {
		memset( &BufferInfo, 0x00, sizeof( CONSOLE_SCREEN_BUFFER_INFO ) );
		GetConsoleScreenBufferInfo( ConsoleHandle, &BufferInfo );
		NumLines		= BufferInfo.dwSize.Y;


		NewLine = FindNextMark( ConsoleHandle, &BufferInfo, Up );


		if ( NewLine >= 0 )
			if ( ScrollToLine( ConsoleHandle, &BufferInfo, NewLine ) )
				Success = TRUE;

		CloseHandle( ConsoleHandle );
	}

	if ( !Success )
		ErrorBeep();
	
	return 0;
}



int ScrollyUp( KEYINFO24 *ki )
{
	ScrollToMark( ki->nHomeRow, TRUE );

	ki->nKey = 0;
	if ( NewKeyHook )
		ki->pszKey[0] = '\0';

	return 0;
}



int ScrollyDown( KEYINFO24 *ki )
{
	ScrollToMark( ki->nHomeRow, FALSE );

	ki->nKey = 0;
	if ( NewKeyHook )
		ki->pszKey[0] = '\0';

	return 0;
}



int RedrawScrollyMarks()
{
	HANDLE ConsoleHandle		= NULL;
	CONSOLE_SCREEN_BUFFER_INFO BufferInfo;

	const int ChunkSize			= 512;
	CHAR_INFO ReadBuffer[ChunkSize][2];
	COORD BufferSize;
	COORD ReadCoord;
	SMALL_RECT ReadRegion;
	WORD Attr					= 0;
	BOOL Dirty					= FALSE;

	int k						= 0;


	ConsoleHandle = CreateFile( L"CONOUT$", GENERIC_READ | GENERIC_WRITE, FILE_SHARE_READ | FILE_SHARE_WRITE, NULL, OPEN_EXISTING, 0, NULL );

	if ( !ConsoleHandle || ( ConsoleHandle == INVALID_HANDLE_VALUE ) )
		return -1;

	memset( &BufferInfo, 0x00, sizeof( CONSOLE_SCREEN_BUFFER_INFO ) );
	if ( !GetConsoleScreenBufferInfo( ConsoleHandle, &BufferInfo ) )
		goto Fail;


	BufferSize.X	= 2;
	BufferSize.Y	= ChunkSize;
	ReadCoord.X		= 0;
	ReadCoord.Y		= 0;


	for ( int row = 0; row < BufferInfo.dwSize.Y; row ++ ) {

		if ( ( row % ChunkSize ) == 0 ) {
			ReadRegion.Top		= row;
			ReadRegion.Bottom	= row + ChunkSize - 1;
			ReadRegion.Left		= 0;
			ReadRegion.Right	= 1;

			if ( ReadRegion.Bottom >= BufferInfo.dwSize.Y )
				ReadRegion.Bottom = BufferInfo.dwSize.Y - 1;

			if ( !ReadConsoleOutput( ConsoleHandle, (CHAR_INFO*) ReadBuffer, BufferSize, ReadCoord, &ReadRegion ) )
				goto Fail;

			Dirty = FALSE;
		}

		Attr = ReadBuffer[row % ChunkSize][0].Attributes;

		if ( Attr & MarkValue ) {

			if ( ScrollyFlags & SCROLLY_FLAG_VISIBLE_MARKS )
				ReadBuffer[row % ChunkSize][0].Attributes = ( Attr & CONSOLE_HIGH_ATTRIBUTES ) | MarkValue | SCROLLY_HIGHLIGHT_1;
			else
				ReadBuffer[row % ChunkSize][0].Attributes = ( Attr & CONSOLE_HIGH_ATTRIBUTES ) | MarkValue | ( ReadBuffer[row % ChunkSize][1].Attributes & CONSOLE_LOW_ATTRIBUTES );

			Dirty = TRUE;
			k++;
		}

		if ( Dirty && ( row == ReadRegion.Bottom ) ) {
			if ( !WriteConsoleOutput( ConsoleHandle, (CHAR_INFO*) ReadBuffer, BufferSize, ReadCoord, &ReadRegion ) )
				goto Fail;

			Dirty = FALSE;
		}
	}


	CloseHandle( ConsoleHandle );
	return k;


Fail:

	CloseHandle( ConsoleHandle );
	return -1;
}


#endif	//	SCROLLY_VERSION


/*

	1.0.0.0		2022-05-17

	1.1.0.0		2022-05-27
				Added RedrawScrollyMarks() to redraw all marks currently in the buffer,
				making them all visible or invisible according to ScrollyFlags.

*/
