
//		SwapArgs.cpp


#ifndef SWAP_ARGS_VER
#define SWAP_ARGS_VER						1.0.1.0


#define VAR_TYPE_NONE						0			//	the main command line
#define VAR_TYPE_FUNCTION					1			//	%@FUNC[arg,arg...]
#define VAR_TYPE_ARRAY						2			//	%ARRAY[i,j...]
#define VAR_TYPE_EVAL						3			//	%((expr))
#define VAR_TYPE_CONDITIONAL				4			//	%[[condition]]
#define VAR_TYPE_INDIRECT					5			//	%[varname]
#define VAR_TYPE_MAX						5



struct _ParseVars {
	_ParseVars *Next;			//	next structure in linked list
	_ParseVars *Parent;			//	parent of this structure; NULL if this is the root
	BOOL Quote;					//	between double quotes
	unsigned int Type;			//	0 - command line, 1 - function args list, 2 - array indices
	wchar_t *SaveArgStart;		//	saves the ArgStart var when entering a nested level
	wchar_t *SaveArgEnd;		//	saves the ArgEnd var when entering a nested level
	wchar_t *OpenBracket;		//	start of the current args list; NULL if this is the root
	wchar_t *CloseBracket;		//	end of the current args list; NULL if this is the root
	wchar_t *Arg1;				//	start of the first arg to swap
	size_t Arg1Len;				//	length of Arg1
	wchar_t *Arg2;				//	start of the second arg to swap
	size_t Arg2Len;				//	length of Arg2
};




int SwapArgs_FreeList( _ParseVars *First )
{
	if ( !First )
		return 0;

	_ParseVars *This		= First;
	_ParseVars *Next		= NULL;
	int k					= 0;


	do {
		Next = This->Next;
		free( This );
		This = Next;
		k++;
	}
	while ( This );

	return k;
}



_ParseVars *NewParseVars( unsigned int Type, _ParseVars *Parent, wchar_t *OpenBracket )
{
	_ParseVars *New		= (_ParseVars*) malloc( sizeof( _ParseVars ) );


	if ( !New )
		return NULL;

	memset( New, 0x00, sizeof( _ParseVars ) );

	New->Type			= Type;
	New->Parent			= Parent;
	New->OpenBracket	= OpenBracket;
	return New;
}



BOOL NeedSpace( wchar_t a, wchar_t b )
{
	if ( !iswalnum( a ) && ( a != '%' ) )
		return FALSE;

	if ( iswgraph( b ) )
		return TRUE;

	return FALSE;
}



wchar_t SwapArgs_GetEscapeChar()
{
	wchar_t temp[SHORT_BUF_LEN]		= L"%@option[EscapeChar]";


	ExpandVariables( temp, 1 );

	if ( temp[0] && ( temp[1] == '\0' ) )
		if ( !iswalnum( temp[0] ) )
			return temp[0];

	return '^';
}



BOOL SwapArgs_GetUnixPaths()
{
	wchar_t temp[SHORT_BUF_LEN]		= L"%@option[UnixPaths]";


	ExpandVariables( temp, 1 );

	if ( towupper( temp[0] == 'Y' ) )
		return TRUE;

	return FALSE;
}



int SwapArgs( LPKEYINFO24 ki )
{
	wchar_t NewCmdLine[MAX_CMD_LINE_LEN]			= L"";

	wchar_t *ch				= ki->pszLine;
	wchar_t *v				= NULL;

	wchar_t EscapeChar		= SwapArgs_GetEscapeChar();
	BOOL UnixPaths			= SwapArgs_GetUnixPaths();

	BOOL Escape				= FALSE;
	BOOL Quote				= FALSE;
	BOOL Done				= FALSE;
	BOOL EndFunction		= FALSE;

	unsigned int VarType	= VAR_TYPE_NONE;
	unsigned int Depth		= 0;
	unsigned int Deepest	= 0;

	wchar_t *ArgStart		= NULL;
	wchar_t *ArgEnd			= NULL;

	_ParseVars CmdLineVars;
	_ParseVars *This			= &CmdLineVars;
	_ParseVars *Best			= &CmdLineVars;
	_ParseVars *Last			= NULL;
	_ParseVars *New				= NULL;

	size_t len					= 0;
	size_t CursorPos			= 0;


	if ( ( wcslen( ki->pszLine ) + 4 ) >= MAX_CMD_LINE_LEN )	//	4 is the max number of spaces that might be added
		goto Error;


	memset( &CmdLineVars, 0x00, sizeof( _ParseVars ) );


	//		Skip past any leading spaces :

	while ( iswspace( *ch ) )
		ch++;


	//		Skip past the command name :

	while ( *ch ) {

		if ( *ch == EscapeChar )
			Escape = !Escape;

		if ( !Escape && ( *ch == QUOTE ) )
			Quote = !Quote;
		else {
			if ( !Quote && iswspace( *ch ) )
				break;

			if ( !Quote && !UnixPaths && ( *ch == '/' ) )
				break;
		}

		ch++;
	}

	while ( iswspace( *ch ) )
		ch++;

	if ( *ch == '\0' )
		goto Error;


	//		Main loop :

	do {

		//		Handle escape character and double quotes :

		if ( *ch == EscapeChar )
			Escape = !Escape;

		if ( !Escape && ( *ch == QUOTE ) )
			Quote = !Quote;


		//		Check for end of argument :

		if ( ArgStart && ( *ch == '\0' ) ) {
			ArgEnd = ch;
			EndFunction = TRUE;
		}
		else if ( ArgStart && !Quote && !Escape ) {

			switch ( This->Type ) {

			case VAR_TYPE_CONDITIONAL :

				if ( iswspace( *ch ) )
					ArgEnd = ch;

				if ( *ch == ']' ) {
					if ( ch[1] == ']' ) {
						ArgEnd = ch;
						EndFunction = TRUE;
						ch++;
					}
				}

				break;

			case VAR_TYPE_EVAL :

				if ( iswspace( *ch ) )
					ArgEnd = ch;

				if ( *ch == ')' ) {
					if ( ch[1] == ')' ) {
						ArgEnd = ch;
						EndFunction = TRUE;
						ch++;
					}
				}

				break;

			case VAR_TYPE_FUNCTION :
			case VAR_TYPE_ARRAY :

				if ( *ch == ',' )
					ArgEnd = ch;

				if ( *ch == ']' ) {
					ArgEnd = ch;
					EndFunction = TRUE;
				}

				break;

			case VAR_TYPE_INDIRECT :

				if ( iswspace( *ch ) )
					ArgEnd = ch;

				if ( *ch == ']' ) {
					ArgEnd = ch;
					EndFunction = TRUE;
				}

				break;

			case VAR_TYPE_NONE :
			default :

				if ( iswspace( *ch ) )
					ArgEnd = ch;

				if ( ( *ch == '/' ) && ( *ArgStart == '/' ) )
					if ( ArgStart[1] != '/' )
						ArgEnd = ch;
			}
		}

		if ( ArgEnd ) {
			if ( ( ArgStart <= ki->pszCurrent ) || ( This->Arg1 == NULL ) ) {
				This->Arg1		= This->Arg2;
				This->Arg1Len	= This->Arg2Len;
				This->Arg2		= ArgStart;
				This->Arg2Len	= ArgEnd - ArgStart;
			}

			ArgStart = ArgEnd = NULL;
		}


		//		Check for start of argument :

		if ( !ArgStart && *ch && !iswspace( *ch ) ) {
			ArgStart = ch;

			if ( Depth && ( *ch == ',' ) )
				ArgStart++;
		}


		//		Handle percent sign :

		if ( !Escape && ( *ch == '%' ) ) {

			VarType = VAR_TYPE_NONE;

			/*
			if ( ch[1] == '%' )
				ch++;
			else */
			
			if ( ch[1] == '@' ) {

				if ( isalpha( ch[2] ) ) {
					v = ch + 2;

					while ( __iswcsym( *v ) || ( *v == '$' ) )
						v++;

					if ( ( v[0] == '[' ) && v[1] ) {
						VarType = VAR_TYPE_FUNCTION;
						ch = v;
					}
				}
			}
			else if ( isalpha( ch[1] ) ) {
				v = ch + 1;

				while ( __iswcsym( *v ) || ( *v == '$' ) )
					v++;

				if ( ( v[0] == '[' ) && v[1] ) {
					VarType = VAR_TYPE_ARRAY;
					ch = v;
				}
			}
			else if ( ch[1] == '(' ) {
				if ( ( ch[2] == '(' ) && ch[3] ) {
					ch += 2;
					VarType = VAR_TYPE_EVAL;
				}
			}
			else if ( ch[1] == '[' ) {
				if ( ( ch[2] == '[' ) && ch[3] ) {
					ch += 2;
					VarType = VAR_TYPE_CONDITIONAL;
				}
				else if ( ch[2] ) {
					ch++;
					VarType = VAR_TYPE_INDIRECT;
				}
			}
			else if ( ch[1] == '=' )
				Escape = TRUE;


			if ( VarType ) {

				//	enter new function args list or array indices list :

				This->Quote			= Quote;
				This->SaveArgStart	= ArgStart;
				This->SaveArgEnd	= ArgEnd;

				Escape		= FALSE;
				Quote		= FALSE;
				ArgStart	= NULL;
				ArgEnd		= NULL;

				Depth++;

				New = NewParseVars( VarType, This, ch );
				if ( !New )
					goto Error;

				if ( Last )
					Last->Next = New;
				else
					CmdLineVars.Next = New;

				This = Last = New;
				VarType = VAR_TYPE_NONE;
			}
		}


		//		Other checks :

		if ( *ch != EscapeChar )
			Escape = FALSE;

		if ( *ch == '\0' )
			Done = EndFunction = TRUE;


		//		Handle end of function args list :

		if ( Depth && EndFunction ) {
			This->CloseBracket = ch;

			if ( ( Depth > Deepest ) && This->Arg1 && ( ki->pszCurrent > This->OpenBracket ) && ( ki->pszCurrent <= This->CloseBracket ) ) {
				Deepest		= Depth;
				Best		= This;
			}

			This = This->Parent;
			Depth--;

			Quote		= This->Quote;
			ArgStart	= This->SaveArgStart;
			ArgEnd		= This->SaveArgEnd;
			EndFunction	= FALSE;
		}

		//		End of main loop :

		ch++;

	} while ( !Done );


	if ( ( Best->Arg1 == NULL ) || ( Best->Arg2 == NULL ) )
		goto Error;


	//		Append command line before arg1 :

	if ( Best->Arg1 > ki->pszLine )
		wcsncpy_s( NewCmdLine, MAX_CMD_LINE_LEN, ki->pszLine, Best->Arg1 - ki->pszLine );

	//		Append arg2 :

	if ( !Deepest && NeedSpace( LastChar( NewCmdLine ), Best->Arg2Len ? *(Best->Arg2) : '\0' ) )
		wcscat_s( NewCmdLine, MAX_CMD_LINE_LEN, L" " );

	if ( Best->Arg2Len )
		wcsncat_s( NewCmdLine, MAX_CMD_LINE_LEN, Best->Arg2, Best->Arg2Len );

	//		Append command line between arg1 and arg2 :

	ch = Best->Arg1 + Best->Arg1Len;

	if ( ch < Best->Arg2 ) {
		len = Best->Arg2 - ch;

		if ( !Deepest && NeedSpace( LastChar( NewCmdLine ), *ch ) )
			wcscat_s( NewCmdLine, MAX_CMD_LINE_LEN, L" " );

		wcsncat_s( NewCmdLine, MAX_CMD_LINE_LEN, ch, len );
	}

	//		Append arg1 :

	if ( !Deepest && NeedSpace( LastChar( NewCmdLine ), Best->Arg1Len ? *(Best->Arg1) : '\0' ) )
		wcscat_s( NewCmdLine, MAX_CMD_LINE_LEN, L" " );

	if ( Best->Arg1Len )
		wcsncat_s( NewCmdLine, MAX_CMD_LINE_LEN, Best->Arg1, Best->Arg1Len );

	CursorPos = wcslen( NewCmdLine );

	//		Append remainder of original command line :

	ch = Best->Arg2 + Best->Arg2Len;

	if ( *ch ) {
		if ( !Deepest && NeedSpace( LastChar( NewCmdLine ), *ch ) )
			wcscat_s( NewCmdLine, MAX_CMD_LINE_LEN, L" " );

		wcscat_s( NewCmdLine, MAX_CMD_LINE_LEN, ch );
	}

	SwapArgs_FreeList( CmdLineVars.Next );

	//	ClearEditLine( ki->pszLine );
	wcscpy_s( ki->pszLine, MAX_CMD_LINE_LEN, NewCmdLine );
	ki->pszCurrent = ki->pszLine + CursorPos;
	ki->fRedraw = TRUE;

	ki->nKey = 0;
	if ( NewKeyHook )
		ki->pszKey[0] = '\0';

	return 0;


Error:

	SwapArgs_FreeList( CmdLineVars.Next );

	ki->nKey = 0;
	if ( NewKeyHook )
		ki->pszKey[0] = '\0';

	ErrorBeep();
	return 0;
}


#endif	//	SWAP_ARGS_VER


/*

	1.0.0.0		2022-05-09

	1.0.1.0		2022-05-10
				Removed last vestiges of configuration options.

*/
