
//		SwapClipWord.cpp


#ifndef SWAP_CLIP_WORD_VER
#define SWAP_CLIP_WORD_VER			1.0.0.0


#ifndef CR
#define CR							0x0d
#endif	//	CR

#ifndef LF
#define LF							0x0a
#endif	//	LF

#ifndef QUOTE
#define QUOTE						0x22
#endif	//	QUOTE



BOOL SCW_UnbalancedQuotes( LPTSTR String )
{
	if ( !String )
		return FALSE;


	int k			= 0;


	for ( wchar_t *ch = String; *ch; ch++ )
		if ( *ch == QUOTE )
			k++;

	return k & 1;
}



int SCW_GetClipWord( LPTSTR outWord, size_t buflen )
{
	if ( !outWord || ( buflen < 128 ) )
		return -1;



	HANDLE ClipDataHandle		= NULL;
	wchar_t *ClipData			= NULL;

	BOOL quote					= FALSE;
	size_t j					= 0;


	outWord[0] = '\0';


	if ( OpenClipboard( GetConsoleWindow() ) ) {
		ClipDataHandle = GetClipboardData( CF_UNICODETEXT );
		
		if ( ClipDataHandle ) {
			ClipData = (wchar_t*) GlobalLock( ClipDataHandle );

			if ( ClipData ) {

				for ( wchar_t *ch = ClipData; *ch; ch++ ) {

					if ( j && ( ( *ch == CR ) || ( *ch == LF ) ) )
						break;

					if ( *ch == QUOTE )
						quote = !quote;

					if ( !quote && iswspace( *ch ) ) {
						if ( j )
							break;
						else
							continue;
					}

					if ( j >= ( buflen - 1 ) ) {
						GlobalUnlock( ClipDataHandle );
						CloseClipboard();
						return 0;
					}

					outWord[j++] = *ch;
					outWord[j] = '\0';
				}

				GlobalUnlock( ClipDataHandle );
			}
		}

		CloseClipboard();
	}

	if ( j ) {
		if ( SCW_UnbalancedQuotes( outWord ) && ( j < ( buflen - 1 ) ) ) {
			outWord[j++] = QUOTE;
			outWord[j] = '\0';
		}
	}


	if ( j )
		return 1;
	else
		return 0;
}



int SwapClipWord( KEYINFO24 *ki )
{
	wchar_t CursorWord[MAX_ARG_LEN]			= L"";
	wchar_t ClipWord[MAX_ARG_LEN]			= L"";

	size_t WordStart			= 0;
	size_t WordLen				= 0;

	HGLOBAL MemHandle			= NULL;
	wchar_t *Buffer				= NULL;
	int rv						= 1;
	

	if ( !FindWordAtCursor( ki, &WordStart, &WordLen, CursorWord, MAX_ARG_LEN ) )
		goto Error;

	SCW_GetClipWord( ClipWord, MAX_ARG_LEN );

	rv = 1;

	MemHandle = GlobalAlloc( GHND, ( WordLen + 1 ) * sizeof( wchar_t ) );

	if ( MemHandle ) {
		Buffer = (wchar_t*) GlobalLock( MemHandle );
		if ( Buffer ) {
			wcscpy_s( Buffer, WordLen + 1, CursorWord );
			GlobalUnlock( MemHandle );
			rv = 0;
		}
	}

	if ( rv )
		goto Error;

	rv = 1;

	if ( OpenClipboard( GetConsoleWindow() ) ) {
		if ( EmptyClipboard() )
			if ( SetClipboardData( CF_UNICODETEXT, MemHandle ) )
				rv = 0;

		CloseClipboard();
	}

	if ( rv )
		goto Error;


	CmdLineDelete( ki, WordStart, WordLen );
	CmdLineInsert( ki, ClipWord, WordStart );

	ki->nKey = 0;
	if ( NewKeyHook )
		ki->pszKey[0] = '\0';

	return 0;


Error :

	if ( MemHandle )
		GlobalFree( MemHandle );

	ErrorBeep();

	ki->nKey = 0;
	if ( NewKeyHook )
		ki->pszKey[0] = '\0';

	return 0;
}


#endif	//	SWAP_CLIP_WORD_VER
